"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import logging
import os
import pathlib as pl
import subprocess
from decimal import Decimal as dec

from PySide6.QtWidgets import QApplication, QFileDialog, QInputDialog, QMessageBox

from . import config as cfg
from . import db_functions, dialog, project_functions, select_observations, select_subj_behav
from . import utilities as util


def events_snapshots(self):
    """
    create snapshots corresponding to coded events
    if observations are from media file and media files have video
    """

    _, selected_observations = select_observations.select_observations2(
        self, cfg.MULTIPLE, windows_title="Select observations for snapshots"
    )
    if not selected_observations:
        return

    # check if obs are MEDIA
    live_images_obs_list = []
    for obs_id in selected_observations:
        if self.pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] in [cfg.LIVE, cfg.IMAGES]:
            live_images_obs_list.append(obs_id)

    if live_images_obs_list:
        out = "The following observations are live observations or observation from images and will be removed from analysis<br><br>"
        out += "<br>".join(live_images_obs_list)
        results = dialog.Results_dialog()
        results.setWindowTitle(cfg.programName)
        results.ptText.setReadOnly(True)
        results.ptText.appendHtml(out)
        results.pbSave.setVisible(False)
        results.pbCancel.setVisible(True)
        if not results.exec_():
            return

    # remove live observations
    selected_observations = [x for x in selected_observations if x not in live_images_obs_list]
    if not selected_observations:
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=dec("NaN"),
        end_coding=dec("NaN"),
        show_include_modifiers=False,
        show_exclude_non_coded_behaviors=False,
        n_observations=len(selected_observations),
    )

    if parameters == {}:
        return

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        return

    ib = dialog.Input_dialog(
        label_caption="Choose parameters",
        elements_list=[
            ("dsb", "Time interval around the events (in seconds)", 0.0, 86400, 1, 0, 3),
            ("il", "Bitmap format", (("JPG - small size / low quality", ""), ("PNG - big size / high quality", ""))),
        ],
        title="Extract frames",
    )
    if not ib.exec_():
        return
    time_interval = util.float2decimal(ib.elements["Time interval around the events (in seconds)"].value())
    if "JPG" in ib.elements["Bitmap format"].currentText():
        frame_bitmap_format = "jpg"
    elif "PNG" in ib.elements["Bitmap format"].currentText():
        frame_bitmap_format = "png"
    else:
        return

    # directory for saving frames
    export_dir = QFileDialog.getExistingDirectory(
        self,
        "Choose a directory to extract events",
        os.path.expanduser("~"),
        options=QFileDialog.ShowDirsOnly,
    )
    if not export_dir:
        return

    cursor = db_functions.load_events_in_db(
        self.pj,
        parameters[cfg.SELECTED_SUBJECTS],
        selected_observations,
        parameters[cfg.SELECTED_BEHAVIORS],
        time_interval=cfg.TIME_FULL_OBS,
    )

    for obs_id in selected_observations:
        for nplayer in self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE]:
            if not self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer]:
                continue
            duration1 = []  # in seconds
            for mediaFile in self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer]:
                duration1.append(self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.LENGTH][mediaFile])

            for subject in parameters[cfg.SELECTED_SUBJECTS]:
                for behavior in parameters[cfg.SELECTED_BEHAVIORS]:
                    cursor.execute(
                        "SELECT occurence FROM events WHERE observation = ? AND subject = ? AND code = ?",
                        (obs_id, subject, behavior),
                    )
                    rows = [{"occurence": util.float2decimal(r["occurence"])} for r in cursor.fetchall()]

                    behavior_state = project_functions.event_type(behavior, self.pj[cfg.ETHOGRAM])

                    for idx, row in enumerate(rows):
                        mediaFileIdx = [idx1 for idx1, x in enumerate(duration1) if row["occurence"] >= sum(duration1[0:idx1])][-1]

                        # check if media has video
                        flag_no_video = False
                        try:
                            flag_no_video = not self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.HAS_VIDEO][
                                self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                            ]
                        except Exception:
                            flag_no_video = True

                        if flag_no_video:
                            logging.debug(f"Media {self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]} does not have video")
                            flag_no_video = True
                            response = dialog.MessageDialog(
                                cfg.programName,
                                (
                                    "The following media file does not have video.<br>"
                                    f"{self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]}"
                                ),
                                [cfg.OK, "Abort"],
                            )
                            if response == cfg.OK:
                                continue
                            if response == "Abort":
                                return

                        # check FPS
                        mediafile_fps = 0
                        try:
                            if self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.FPS][
                                self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                            ]:
                                mediafile_fps = util.float2decimal(
                                    self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.FPS][
                                        self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                                    ]
                                )
                        except Exception:
                            mediafile_fps = 0

                        if not mediafile_fps:
                            logging.debug(f"FPS not found for {self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]}")
                            response = dialog.MessageDialog(
                                cfg.programName,
                                (
                                    "The FPS was not found for the following media file:<br>"
                                    f"{self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]}"
                                ),
                                [cfg.OK, "Abort"],
                            )
                            if response == cfg.OK:
                                continue
                            if response == "Abort":
                                return

                        global_start = dec("0.000") if row["occurence"] < time_interval else round(row["occurence"] - time_interval, 3)
                        start = round(
                            row["occurence"]
                            - time_interval
                            - util.float2decimal(sum(duration1[0:mediaFileIdx]))
                            - self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET],
                            3,
                        )
                        if start < time_interval:
                            start = dec("0.000")

                        if behavior_state in cfg.POINT_EVENT_TYPES:
                            media_path = project_functions.full_path(
                                self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx],
                                self.projectFileName,
                            )

                            vframes = 1 if not time_interval else int(mediafile_fps * time_interval * 2)
                            if vframes == 0:
                                vframes = 1

                        if behavior_state in cfg.STATE_EVENT_TYPES:
                            if idx % 2 == 0:
                                # check if stop is on same media file
                                if (
                                    mediaFileIdx
                                    != [idx1 for idx1, x in enumerate(duration1) if rows[idx + 1]["occurence"] >= sum(duration1[0:idx1])][
                                        -1
                                    ]
                                ):
                                    response = dialog.MessageDialog(
                                        cfg.programName,
                                        (
                                            "The event extends on 2 video. "
                                            "At the moment it no possible to extract frames "
                                            "for this type of event.<br>"
                                        ),
                                        [cfg.OK, "Abort"],
                                    )
                                    if response == cfg.OK:
                                        continue
                                    if response == "Abort":
                                        return

                                # globalStop = round(rows[idx + 1]["occurence"] + time_interval, 3)

                                stop = round(
                                    rows[idx + 1]["occurence"]
                                    + time_interval
                                    - util.float2decimal(sum(duration1[0:mediaFileIdx]))
                                    - self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET],
                                    3,
                                )

                                # check if start after length of media
                                try:
                                    if (
                                        start
                                        > self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.LENGTH][
                                            self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                                        ]
                                    ):
                                        continue
                                except Exception:
                                    continue

                                media_path = project_functions.full_path(
                                    self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx],
                                    self.projectFileName,
                                )

                                vframes = int((stop - start) * mediafile_fps + time_interval * mediafile_fps * 2)
                                if vframes == 0:
                                    vframes = 1

                            else:
                                continue

                        ffmpeg_command = (
                            f'"{self.ffmpeg_bin}" '
                            f"-ss {start:.3f} "
                            f'-i "{media_path}" '
                            f"-vframes {vframes} "
                            f'"{export_dir}{os.sep}'
                            f"{util.safeFileName(obs_id).replace(' ', '-')}"
                            f"_PLAYER{nplayer}"
                            f"_{util.safeFileName(subject).replace(' ', '-')}"
                            f"_{util.safeFileName(behavior).replace(' ', '-')}"
                            f'_{global_start:.3f}_%08d.{frame_bitmap_format}"'
                        )

                        logging.debug(f"ffmpeg command: {ffmpeg_command}")

                        p = subprocess.Popen(ffmpeg_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                        out, _ = p.communicate()

    self.statusbar.showMessage(f"Frames extracted in {export_dir}", 0)


def extract_events(self):
    """
    extract sub-sequences from media files corresponding to coded events with FFmpeg
    in case of point event, from -n to +n seconds are extracted (n is asked to user)
    """

    _, selected_observations = select_observations.select_observations2(
        self, cfg.MULTIPLE, windows_title="Select observations for extracting events"
    )
    if not selected_observations:
        return

    # check if obs are MEDIA
    live_images_obs_list = []
    for obs_id in selected_observations:
        if self.pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] in [cfg.LIVE, cfg.IMAGES]:
            live_images_obs_list.append(obs_id)

    if live_images_obs_list:
        out = "The following observations are live observations or observation from pictures and will be removed from analysis<br><br>"
        out += "<br>".join(live_images_obs_list)
        results = dialog.Results_dialog()
        results.setWindowTitle(cfg.programName)
        results.ptText.setReadOnly(True)
        results.ptText.appendHtml(out)
        results.pbSave.setVisible(False)
        results.pbCancel.setVisible(True)
        if results.exec_():
            # remove live observations
            selected_observations = [x for x in selected_observations if x not in live_images_obs_list]
            if not selected_observations:
                return
        else:
            return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=dec("NaN"),
        end_coding=dec("NaN"),
        show_include_modifiers=False,
        show_exclude_non_coded_behaviors=False,
    )
    if parameters == {}:
        return

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    # Ask for time interval around the event
    while True:
        text, ok = QInputDialog.getDouble(self, "Time interval around the events", "Time (in seconds):", 0.0, 0.0, 86400, 1)
        if not ok:
            return
        try:
            timeOffset = util.float2decimal(text)
            break
        except Exception:
            QMessageBox.warning(self, cfg.programName, f"<b>{text}</b> is not recognized as time")

    # ask for video / audio extraction
    items_to_extract, ok = QInputDialog.getItem(
        self, "Tracks to extract", "Tracks", ("Video and audio", "Only video", "Only audio"), 0, False
    )
    if not ok:
        return

    export_dir = QFileDialog.getExistingDirectory(
        self,
        "Choose a directory to extract events",
        os.path.expanduser("~"),
        options=QFileDialog.ShowDirsOnly,
    )
    if not export_dir:
        return

    cursor = db_functions.load_events_in_db(
        self.pj,
        parameters[cfg.SELECTED_SUBJECTS],
        selected_observations,
        parameters[cfg.SELECTED_BEHAVIORS],
        time_interval=cfg.TIME_FULL_OBS,
    )

    self.statusBar().showMessage("Extracting sequences from media files")
    QApplication.processEvents()

    ffmpeg_extract_command: str = '"{ffmpeg_bin}" -ss {start} -i "{input_}" -y -t {duration} {codecs} '
    mem_command: str = ""
    for obs_id in selected_observations:
        for nplayer in self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE]:
            if not self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer]:
                continue

            duration1 = []  # in seconds
            for mediaFile in self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer]:
                duration1.append(self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.LENGTH][mediaFile])

            for subject in parameters[cfg.SELECTED_SUBJECTS]:
                for behavior in parameters[cfg.SELECTED_BEHAVIORS]:
                    cursor.execute(
                        "SELECT occurence FROM events WHERE observation = ? AND subject = ? AND code = ?",
                        (obs_id, subject, behavior),
                    )
                    rows = [{"occurence": util.float2decimal(r["occurence"])} for r in cursor.fetchall()]

                    behavior_state = project_functions.event_type(behavior, self.pj[cfg.ETHOGRAM])
                    if behavior_state in cfg.STATE_EVENT_TYPES and len(rows) % 2:  # unpaired events
                        continue

                    for idx, row in enumerate(rows):
                        mediaFileIdx = [idx1 for idx1, x in enumerate(duration1) if row["occurence"] >= sum(duration1[0:idx1])][-1]

                        if "VIDEO" in items_to_extract.upper():
                            # check if media has video
                            has_video = False
                            try:
                                has_video = self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.HAS_VIDEO][
                                    self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                                ]
                            except Exception:
                                has_video = False
                            if not has_video:
                                if (
                                    dialog.MessageDialog(
                                        cfg.programName,
                                        f"The media file {self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]} does not have a video stream",
                                        ["Continue", "Abort"],
                                    )
                                    == "Abort"
                                ):
                                    return
                                else:
                                    continue

                            new_extension = ".mp4"
                            if items_to_extract == "Only video":
                                codecs = "-an"
                            else:
                                codecs = ""

                        if items_to_extract == "Only audio":
                            # check if media has audio
                            has_audio = False
                            try:
                                has_audio = self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.HAS_AUDIO][
                                    self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                                ]
                            except Exception:
                                has_audio = False
                            if not has_audio:
                                if (
                                    dialog.MessageDialog(
                                        cfg.programName,
                                        f"The media file {self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]} does not have an audio stream",
                                        ["Continue", "Abort"],
                                    )
                                    == "Abort"
                                ):
                                    return
                                else:
                                    continue

                            new_extension = ".wav"
                            codecs = "-vn"

                        if behavior_state in cfg.POINT_EVENT_TYPES:
                            globalStart = dec("0.000") if row["occurence"] < timeOffset else round(row["occurence"] - timeOffset, 3)
                            start = round(
                                row["occurence"]
                                - (timeOffset if timeOffset else 1)  # if time offset is not set default = 1 s
                                - util.float2decimal(sum(duration1[0:mediaFileIdx]))
                                - self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET],
                                3,
                            )
                            if start < timeOffset:
                                start = dec("0.000")

                            globalStop = round(row["occurence"] + timeOffset, 3)

                            stop = round(
                                row["occurence"]
                                + (timeOffset if timeOffset else 1)  # if time offset is not set default = 1 s
                                - util.float2decimal(sum(duration1[0:mediaFileIdx]))
                                - self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET],
                                3,
                            )

                        if behavior_state in cfg.STATE_EVENT_TYPES:
                            if idx % 2 == 0:
                                # check if stop is on same media file
                                if (
                                    mediaFileIdx
                                    != [idx1 for idx1, x in enumerate(duration1) if rows[idx + 1]["occurence"] >= sum(duration1[0:idx1])][
                                        -1
                                    ]
                                ):
                                    response = dialog.MessageDialog(
                                        cfg.programName,
                                        (
                                            "The event extends on 2 successive video. "
                                            " At the moment it is not possible to extract this type of event.<br>"
                                        ),
                                        [cfg.OK, "Abort"],
                                    )
                                    if response == cfg.OK:
                                        continue
                                    if response == "Abort":
                                        return

                                globalStart = dec("0.000") if row["occurence"] < timeOffset else round(row["occurence"] - timeOffset, 3)
                                start = round(
                                    row["occurence"]
                                    - timeOffset
                                    - util.float2decimal(sum(duration1[0:mediaFileIdx]))
                                    - self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET],
                                    3,
                                )
                                if start < timeOffset:
                                    start = dec("0.000")

                                globalStop = round(rows[idx + 1]["occurence"] + timeOffset, 3)

                                stop = round(
                                    rows[idx + 1]["occurence"]
                                    + timeOffset
                                    - util.float2decimal(sum(duration1[0:mediaFileIdx]))
                                    - self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET],
                                    3,
                                )

                                # check if start after length of media
                                if (
                                    start
                                    > self.pj[cfg.OBSERVATIONS][obs_id][cfg.MEDIA_INFO][cfg.LENGTH][
                                        self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx]
                                    ]
                                ):
                                    continue

                            else:
                                continue

                        new_file_name = pl.Path(export_dir) / pl.Path(
                            (
                                f"{util.safeFileName(obs_id).replace(' ', '-')}_"
                                f"PLAYER{nplayer}_"
                                f"{util.safeFileName(subject).replace(' ', '-')}_"
                                f"{util.safeFileName(behavior)}_"
                                f"{globalStart}-{globalStop}"
                                f"{new_extension}"
                            )
                        )  # .with_suffix(new_extension)

                        if new_file_name.is_file():
                            if mem_command not in (cfg.OVERWRITE_ALL, cfg.SKIP_ALL):
                                mem_command = dialog.MessageDialog(
                                    cfg.programName,
                                    f"The file <b>{new_file_name}</b> already exists.",
                                    [cfg.OVERWRITE, cfg.OVERWRITE_ALL, cfg.SKIP, cfg.SKIP_ALL, cfg.CANCEL],
                                )
                            if mem_command == cfg.CANCEL:
                                return
                            if "SKIP" in mem_command.upper():
                                continue

                        ffmpeg_command = ffmpeg_extract_command.format(
                            ffmpeg_bin=self.ffmpeg_bin,
                            input_=project_functions.full_path(
                                self.pj[cfg.OBSERVATIONS][obs_id][cfg.FILE][nplayer][mediaFileIdx],
                                self.projectFileName,
                            ),
                            start=start,
                            duration=stop - start,
                            codecs=codecs,
                        )

                        logging.debug(f'ffmpeg command: {ffmpeg_command} "{new_file_name}"')

                        p = subprocess.Popen(
                            f'{ffmpeg_command} "{new_file_name}"',
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            shell=True,
                        )
                        out, _ = p.communicate()

    self.statusbar.showMessage(f"Media sequences extracted in {export_dir}", 0)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import json
import binascii

from PySide6.QtGui import QMouseEvent, QPixmap, QPolygonF, QColor, QBrush, QPen
from PySide6.QtCore import Qt, Signal, QEvent, QPoint
from PySide6.QtWidgets import (
    QLabel,
    QHBoxLayout,
    QGraphicsView,
    QGraphicsScene,
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QPushButton,
    QGraphicsPixmapItem,
    QGraphicsPolygonItem,
    QMessageBox,
    QInputDialog,
    QApplication,
)

from . import config as cfg

codeSeparator: str = ","
penWidth: int = 0
penStyle = Qt.NoPen


class BehaviorsCodingMapWindowClass(QWidget):
    class View(QGraphicsView):
        mousePress = Signal(QMouseEvent)
        mouseMove = Signal(QMouseEvent)

        def eventFilter(self, source, event):
            if event.type() == QEvent.MouseMove:
                self.mouseMove.emit(event)

            if event.type() == QEvent.MouseButtonPress:
                self.mousePress.emit(event)

            return QWidget.eventFilter(self, source, event)

        elList, points = [], []

        def __init__(self, parent):
            QGraphicsView.__init__(self, parent)
            self.setScene(QGraphicsScene(self))
            self.scene().update()

            self.viewport().installEventFilter(self)
            self.setMouseTracking(True)

    clickSignal = Signal(str, list)  # click signal to be sent to mainwindow
    keypressSignal = Signal(QEvent)
    close_signal = Signal(str)

    def __init__(self, behaviors_coding_map, idx=0):
        super(BehaviorsCodingMapWindowClass, self).__init__()

        self.polygonsList2 = []

        self.installEventFilter(self)

        self.codingMap = behaviors_coding_map
        self.idx = idx

        self.setWindowTitle(f"Behaviors coding map: {self.codingMap['name']}")
        Vlayout = QVBoxLayout()

        self.view = self.View(self)
        self.view.mousePress.connect(self.viewMousePressEvent)
        self.view.mouseMove.connect(self.mouse_move_event)

        Vlayout.addWidget(self.view)

        hBoxLayout1 = QHBoxLayout()

        self.label = QLabel("Behavior(s)")
        hBoxLayout1.addWidget(self.label)

        self.leareaCode = QLineEdit(self)
        hBoxLayout1.addWidget(self.leareaCode)

        self.btClose = QPushButton(cfg.CLOSE)
        self.btClose.clicked.connect(self.close)
        hBoxLayout1.addWidget(self.btClose)

        Vlayout.addLayout(hBoxLayout1)

        self.setLayout(Vlayout)

        self.loadMap()

    def closeEvent(self, event):
        self.close_signal.emit(self.codingMap["name"])
        event.accept()

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.keypressSignal.emit(event)
            return True
        else:
            return False

    def mouse_move_event(self, event):
        """
        display behavior under mouse position
        """

        self.leareaCode.clear()
        codes = []
        test = self.view.mapToScene(event.pos()).toPoint()
        for areaCode, pg in self.polygonsList2:
            if pg.contains(test):
                codes.append(areaCode)
        self.leareaCode.setText(", ".join(codes))

    def viewMousePressEvent(self, event) -> None:
        """
        insert clicked areas codes
        """

        test = self.view.mapToScene(event.pos()).toPoint()
        to_be_sent: list = []

        for areaCode, pg in self.polygonsList2:
            if pg.contains(test):
                to_be_sent.append(areaCode)

        if to_be_sent:
            self.clickSignal.emit(self.codingMap["name"], to_be_sent)

    def loadMap(self):
        """
        load bitmap from data
        show it in view scene
        """

        pixmap = QPixmap()
        pixmap.loadFromData(binascii.a2b_base64(self.codingMap["bitmap"]))

        self.view.setSceneRect(0, 0, pixmap.size().width(), pixmap.size().height())
        pixItem = QGraphicsPixmapItem(pixmap)
        pixItem.setPos(0, 0)
        self.view.scene().addItem(pixItem)

        for key in self.codingMap["areas"]:
            areaCode = self.codingMap["areas"][key]["code"]
            points = self.codingMap["areas"][key]["geometry"]

            newPolygon = QPolygonF()
            for p in points:
                newPolygon.append(QPoint(p[0], p[1]))

            # draw polygon
            polygon = QGraphicsPolygonItem()
            polygon.setPolygon(newPolygon)
            clr = QColor()
            clr.setRgba(self.codingMap["areas"][key]["color"])
            polygon.setPen(QPen(clr, penWidth, penStyle, Qt.RoundCap, Qt.RoundJoin))
            polygon.setBrush(QBrush(clr, Qt.SolidPattern))

            self.view.scene().addItem(polygon)

            self.polygonsList2.append([areaCode, polygon])


def show_behaviors_coding_map(self):
    """
    show a behavior coding map
    """

    if cfg.BEHAVIORS_CODING_MAP not in self.pj or not self.pj[cfg.BEHAVIORS_CODING_MAP]:
        QMessageBox.warning(self, cfg.programName, "No behaviors coding map found in current project")
        return

    items = [x["name"] for x in self.pj[cfg.BEHAVIORS_CODING_MAP]]
    if len(items) == 1:
        coding_map_name = items[0]
    else:
        item, ok = QInputDialog.getItem(self, "Select a coding map", "list of coding maps", items, 0, False)
        if ok and item:
            coding_map_name = item
        else:
            return

    if self.bcm_dict.get(coding_map_name, None) is not None:
        # if coding_map_name in self.bcm_dict and :
        self.bcm_dict[coding_map_name].show()
    else:
        self.bcm_dict[coding_map_name] = BehaviorsCodingMapWindowClass(
            self.pj[cfg.BEHAVIORS_CODING_MAP][items.index(coding_map_name)], idx=items.index(coding_map_name)
        )

        self.bcm_dict[coding_map_name].clickSignal.connect(self.click_signal_from_behaviors_coding_map)

        # self.bcm_dict[coding_map_name].close_signal.connect(self.close_behaviors_coding_map)

        self.bcm_dict[coding_map_name].resize(cfg.CODING_MAP_RESIZE_W, cfg.CODING_MAP_RESIZE_W)
        self.bcm_dict[coding_map_name].setWindowFlags(Qt.WindowStaysOnTopHint)
        self.bcm_dict[coding_map_name].show()


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    if len(sys.argv) > 1:
        cm = json.loads(open(sys.argv[1]).read())
        codingMapWindow = BehaviorsCodingMapWindowClass(cm)
        codingMapWindow.resize(cfg.CODING_MAP_RESIZE_W, cfg.CODING_MAP_RESIZE_H)
        codingMapWindow.show()
        sys.exit(app.exec_())



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import sys
import subprocess
import platform
import numpy as np
import pandas as pd
import matplotlib

from . import version
from . import config as cfg
from . import utilities as util


from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QMessageBox


def actionAbout_activated(self):
    """
    About dialog
    """

    programs_versions: list = ["MPV media player"]

    mpv_lib_version, mpv_lib_file_path, mpv_api_version = util.mpv_lib_version()
    programs_versions.append(
        (
            f"Library version: {mpv_lib_version} file: {mpv_lib_file_path}\n"
            f"MPV API version: {mpv_api_version}\n"
            f"python-mpv version: {util.python_mpv_script_version()}"
        )
    )

    # ffmpeg
    if self.ffmpeg_bin == "ffmpeg" and sys.platform.startswith("linux"):
        ffmpeg_true_path = subprocess.getoutput("which ffmpeg")
    else:
        ffmpeg_true_path = self.ffmpeg_bin
    programs_versions.extend(
        [
            "\nFFmpeg",
            subprocess.getoutput(f'"{self.ffmpeg_bin}" -version').split("\n")[0],
            f"Path: {ffmpeg_true_path}",
            "https://www.ffmpeg.org",
        ]
    )

    # numpy
    programs_versions.extend(["\nNumpy", f"version {np.__version__}", "https://numpy.org"])

    # matplotlib
    programs_versions.extend(["\nMatplotlib", f"version {matplotlib.__version__}", "https://matplotlib.org"])

    # pandas
    programs_versions.extend(["\nPandas", f"version {pd.__version__}", "https://pandas.pydata.org"])

    # graphviz
    gv_result = subprocess.getoutput("dot -V")

    programs_versions.extend(["\nGraphViz", gv_result if "graphviz" in gv_result else "not installed", "https://www.graphviz.org/"])

    about_dialog = QMessageBox()
    about_dialog.setIconPixmap(QPixmap(":/boris_unito"))

    about_dialog.setWindowTitle(f"About {cfg.programName}")
    about_dialog.setStandardButtons(QMessageBox.Ok)
    about_dialog.setDefaultButton(QMessageBox.Ok)
    about_dialog.setEscapeButton(QMessageBox.Ok)

    about_dialog.setInformativeText(
        (
            f"<b>{cfg.programName}</b> v. {version.__version__} - {version.__version_date__}"
            "<p>Copyright &copy; 2012-2025 Olivier Friard - Marco Gamba<br>"
            "Department of Life Sciences and Systems Biology<br>"
            "University of Torino - Italy<br>"
            "<br>"
            'BORIS is released under the <a href="https://www.gnu.org/copyleft/gpl.html">GNU General Public License</a><br>'
            'See <a href="https://www.boris.unito.it">www.boris.unito.it</a> for more details.<br>'
            "<br>"
            "The authors would like to acknowledge Valentina Matteucci for her precious help."
            "<hr>"
            "How to cite BORIS:<br>"
            "Friard, O. and Gamba, M. (2016), BORIS: a free, versatile open-source event-logging software for video/audio "
            "coding and live observations. Methods Ecol Evol, 7: 1325–1330.<br>"
            '<a href="https://besjournals.onlinelibrary.wiley.com/doi/full/10.1111/2041-210X.12584">DOI:10.1111/2041-210X.12584</a>'
        )
    )
    """
    n = "\n"
    current_system = platform.uname()
    details = (
        f"Operating system: {current_system.system} {current_system.release} {current_system.version} \n"
        f"CPU: {current_system.machine} {current_system.processor}\n\n"
        f"Python {platform.python_version()} ({'64-bit' if sys.maxsize > 2**32 else '32-bit'})"
        f"Qt {qVersion()} - PySide {PySide6.__version__}\n"
    )

    r, memory = util.mem_info()
    if not r:
        details += (
            f"Memory (RAM)  Total: {memory.get('total_memory', 'Not available'):.2f} Mb  "
            f"Free: {memory.get('free_memory', 'Not available'):.2f} Mb\n\n"
        )
    """

    details = util.get_systeminfo()

    details += "\n".join(programs_versions)
    """
    memory_in_use = f"{utilities.rss_memory_used(self.pid)} Mb" if utilities.rss_memory_used(self.pid) != -1 else "Not available"
    percent_memory_in_use = (f"({utilities.rss_memory_percent_used(self.pid):.1f} % of total memory)"
                                if utilities.rss_memory_percent_used(self.pid) != -1
                                else "")
    """
    """
    f"Total memory: {psutil.virtual_memory().total / 1024 / 1024 / 1024:.1f} Gb "
    f"({100 - psutil.virtual_memory().percent :.1f} % available){n}"
    f"Memory in use by BORIS: {memory_in_use} {percent_memory_in_use}{n}{n}"
    """

    about_dialog.setDetailedText(details)

    _ = about_dialog.exec()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import sqlite3
import logging
from typing import Optional, Tuple
from . import config as cfg
from . import project_functions
from . import event_operations


def load_events_in_db(
    pj: dict,
    selected_subjects: list,
    selected_observations: list,
    selected_behaviors: list,
    time_interval: str = cfg.TIME_FULL_OBS,
):
    """
    populate a memory sqlite database with events from selected_observations,
    selected_subjects and selected_behaviors

    Args:
        pj (dict): project dictionary
        selected_observations (list):
        selected_subjects (list):
        selected_behaviors (list):
        time_interval (str): time interval for loading events (cfg.TIME_FULL_OBS / TIME_cfg.EVENTS / TIME_ARBITRARY_INTERVAL)

    Returns:
        database cursor:

    """

    # selected behaviors defined as state event
    state_behaviors_codes = [
        pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE]
        for x in pj[cfg.ETHOGRAM]
        if cfg.STATE in pj[cfg.ETHOGRAM][x][cfg.TYPE].upper() and pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] in selected_behaviors
    ]

    # selected behaviors defined as point event
    """
    point_behaviors_codes = [
        pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE]
        for x in pj[cfg.ETHOGRAM]
        if cfg.POINT in pj[cfg.ETHOGRAM][x][cfg.TYPE].upper()
        and pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] in selected_behaviors
    ]
    """

    db = sqlite3.connect(":memory:", isolation_level=None)

    """
    import os
    os.system("rm /tmp/ramdisk/events.sqlite")
    db = sqlite3.connect("/tmp/ramdisk/events.sqlite", isolation_level=None)
    """

    db.row_factory = sqlite3.Row
    cursor = db.cursor()
    cursor.execute(
        (
            "CREATE TABLE events (observation TEXT, "
            "subject TEXT, "
            "code TEXT, "
            "type TEXT, "
            "modifiers TEXT, "
            "occurence FLOAT, "
            "comment TEXT,"
            "image_index INTEGER,"
            "image_path TEXT)"
        )
    )

    cursor.execute("CREATE INDEX observation_idx ON events(observation)")
    cursor.execute("CREATE INDEX subject_idx ON events(subject)")
    cursor.execute("CREATE INDEX code_idx ON events(code)")
    cursor.execute("CREATE INDEX modifiers_idx ON events(modifiers)")

    for subject_to_analyze in selected_subjects:
        for obs_id in selected_observations:
            for event in pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]:
                if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] in selected_behaviors:
                    # extract time, code, modifier and comment (time:0, subject:1, code:2, modifier:3, comment:4)
                    if (subject_to_analyze == cfg.NO_FOCAL_SUBJECT and event[cfg.EVENT_SUBJECT_FIELD_IDX] == "") or (
                        event[cfg.EVENT_SUBJECT_FIELD_IDX] == subject_to_analyze
                    ):
                        if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] in (cfg.MEDIA, cfg.LIVE):
                            cursor.execute(
                                (
                                    "INSERT INTO events "
                                    "(observation, subject, code, type, modifiers, occurence, comment, image_index) "
                                    "VALUES (?,?,?,?,?,?,?,?)"
                                ),
                                (
                                    obs_id,
                                    cfg.NO_FOCAL_SUBJECT
                                    if event[cfg.EVENT_SUBJECT_FIELD_IDX] == ""
                                    else event[cfg.EVENT_SUBJECT_FIELD_IDX],
                                    event[cfg.EVENT_BEHAVIOR_FIELD_IDX],
                                    cfg.STATE if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] in state_behaviors_codes else cfg.POINT,
                                    event[cfg.EVENT_MODIFIER_FIELD_IDX],
                                    float(event[cfg.EVENT_TIME_FIELD_IDX]) if not event[cfg.EVENT_TIME_FIELD_IDX].is_nan() else None,
                                    event[cfg.EVENT_COMMENT_FIELD_IDX],
                                    # frame index or NA
                                    event_operations.read_event_field(event, pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE], cfg.FRAME_INDEX),
                                ),
                            )

                        if pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE] == cfg.IMAGES:
                            cursor.execute(
                                (
                                    "INSERT INTO events "
                                    "(observation, subject, code, type, modifiers, occurence, comment, image_index, image_path) "
                                    "VALUES (?,?,?,?,?,?,?,?,?)"
                                ),
                                (
                                    obs_id,
                                    cfg.NO_FOCAL_SUBJECT
                                    if event[cfg.EVENT_SUBJECT_FIELD_IDX] == ""
                                    else event[cfg.EVENT_SUBJECT_FIELD_IDX],
                                    event[cfg.EVENT_BEHAVIOR_FIELD_IDX],
                                    cfg.STATE if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] in state_behaviors_codes else cfg.POINT,
                                    event[cfg.EVENT_MODIFIER_FIELD_IDX],
                                    float(event[cfg.EVENT_TIME_FIELD_IDX]) if not event[cfg.EVENT_TIME_FIELD_IDX].is_nan() else None,
                                    event[cfg.EVENT_COMMENT_FIELD_IDX],
                                    event[cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_INDEX]],
                                    event[cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_PATH]],
                                ),
                            )

    db.commit()
    return cursor


def load_aggregated_events_in_db(
    pj: dict, selected_subjects: list, selected_observations: list, selected_behaviors: list
) -> Tuple[bool, str, Optional[sqlite3.Connection]]:
    """
    populate a memory sqlite database with aggregated events from selected_observations, selected_subjects and selected_behaviors

    Args:
        pj (dict): project dictionary
        selected_observations (list):
        selected_subjects (list):
        selected_behaviors (list):

    Returns:
        bool: True if OK else False
        str: error message
        database connector: db connector if bool True else None

    """

    logging.debug("function: load_aggregated_events_in_db")

    # if no observation selected select all
    if not selected_observations:
        selected_observations = sorted([x for x in pj[cfg.OBSERVATIONS]])

    # if no subject selected select all
    if not selected_subjects:
        selected_subjects = sorted([pj[cfg.SUBJECTS][x][cfg.SUBJECT_NAME] for x in pj[cfg.SUBJECTS]] + [cfg.NO_FOCAL_SUBJECT])

    # if no behavior selected select all
    if not selected_behaviors:
        selected_behaviors = sorted([pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in pj[cfg.ETHOGRAM]])

    # check if state events are paired
    out = ""
    for obs_id in selected_observations:
        r, msg = project_functions.check_state_events_obs(obs_id, pj[cfg.ETHOGRAM], pj[cfg.OBSERVATIONS][obs_id], cfg.HHMMSS)
        if not r:
            out += f"Observation: <strong>{obs_id}</strong><br>{msg}<br>"
    if out:
        return False, out, None

    # selected behaviors defined as state event
    state_behaviors_codes = [
        pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE]
        for x in pj[cfg.ETHOGRAM]
        if cfg.STATE in pj[cfg.ETHOGRAM][x][cfg.TYPE].upper() and pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] in selected_behaviors
    ]

    # selected behaviors defined as point event
    point_behaviors_codes = [
        pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE]
        for x in pj[cfg.ETHOGRAM]
        if cfg.POINT in pj[cfg.ETHOGRAM][x][cfg.TYPE].upper() and pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] in selected_behaviors
    ]

    db = sqlite3.connect(":memory:")

    """
    # only for debugging
    import os
    os.system("rm /tmp/ramdisk/aggreg.sqlite")
    db = sqlite3.connect("/tmp/ramdisk/aggreg.sqlite", isolation_level=None)
    """

    db.row_factory = sqlite3.Row
    cursor2 = db.cursor()
    cursor2.execute(
        (
            "CREATE TABLE aggregated_events "
            "(id INTEGER PRIMARY KEY ASC, "
            "observation TEXT, "
            "subject TEXT, "
            "behavior TEXT, "
            "type TEXT, "
            "modifiers TEXT, "
            "start FLOAT, "
            "stop FLOAT, "
            "comment TEXT, "
            "comment_stop TEXT,"
            "image_index_start INTEGER,"
            "image_index_stop INTEGER,"
            "image_path_start TEXT,"
            "image_path_stop TEXT)"
        )
    )

    cursor2.execute("CREATE INDEX observation_idx ON aggregated_events(observation)")
    cursor2.execute("CREATE INDEX subject_idx ON aggregated_events(subject)")
    cursor2.execute("CREATE INDEX behavior_idx ON aggregated_events(behavior)")
    cursor2.execute("CREATE INDEX modifiers_idx ON aggregated_events(modifiers)")

    # too slow! cursor1 = load_events_in_db(pj, selected_subjects, selected_observations, selected_behaviors)

    insert_sql = (
        "INSERT INTO aggregated_events (observation, subject, behavior, type, modifiers, "
        "                               start, stop, comment, comment_stop, "
        "image_index_start, image_index_stop, image_path_start, image_path_stop) "
        "VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)"
    )

    for obs_id in selected_observations:
        cursor1 = load_events_in_db(pj, selected_subjects, [obs_id], selected_behaviors)

        for subject in selected_subjects:
            for behavior in selected_behaviors:
                cursor1.execute(
                    "SELECT DISTINCT modifiers FROM events WHERE subject=? AND code=? ORDER BY modifiers",
                    (
                        subject,
                        behavior,
                    ),
                )

                rows_distinct_modifiers = list(x[0] for x in cursor1.fetchall())

                for distinct_modifiers in rows_distinct_modifiers:
                    cursor1.execute(
                        (
                            "SELECT occurence, comment, image_index, image_path FROM events "
                            "WHERE subject = ? AND code = ? AND modifiers = ? ORDER by occurence"
                        ),
                        (subject, behavior, distinct_modifiers),
                    )
                    rows = list(cursor1.fetchall())

                    for idx, row in enumerate(rows):
                        if behavior in point_behaviors_codes:
                            data = (
                                obs_id,
                                subject,
                                behavior,
                                cfg.POINT,
                                distinct_modifiers,
                                row["occurence"],
                                row["occurence"],
                                row["comment"],
                                "",  # no stop comment for point event
                                row["image_index"],
                                row["image_index"],
                                row["image_path"],
                                row["image_path"],
                            )
                            cursor2.execute(insert_sql, data)

                        if behavior in state_behaviors_codes:
                            if idx % 2 == 0:
                                data = (
                                    obs_id,
                                    subject,
                                    behavior,
                                    cfg.STATE,
                                    distinct_modifiers,
                                    row["occurence"],
                                    rows[idx + 1]["occurence"],
                                    row["comment"],
                                    rows[idx + 1]["comment"],
                                    row["image_index"],
                                    rows[idx + 1]["image_index"],
                                    row["image_path"],
                                    rows[idx + 1]["image_path"],
                                )

                                cursor2.execute(insert_sql, data)

    db.commit()

    return True, "", db



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'converters.ui'
##
## Created by: Qt User Interface Compiler version 6.8.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QHBoxLayout, QHeaderView,
    QLabel, QLineEdit, QPlainTextEdit, QPushButton,
    QSizePolicy, QSpacerItem, QTableWidget, QTableWidgetItem,
    QVBoxLayout, QWidget)

class Ui_converters(object):
    def setupUi(self, converters):
        if not converters.objectName():
            converters.setObjectName(u"converters")
        converters.resize(1029, 530)
        self.verticalLayout = QVBoxLayout(converters)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label_4 = QLabel(converters)
        self.label_4.setObjectName(u"label_4")

        self.verticalLayout.addWidget(self.label_4)

        self.tw_converters = QTableWidget(converters)
        if (self.tw_converters.columnCount() < 3):
            self.tw_converters.setColumnCount(3)
        __qtablewidgetitem = QTableWidgetItem()
        self.tw_converters.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.tw_converters.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.tw_converters.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        self.tw_converters.setObjectName(u"tw_converters")
        self.tw_converters.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.tw_converters.setSelectionMode(QAbstractItemView.SingleSelection)
        self.tw_converters.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.tw_converters.setSortingEnabled(True)

        self.verticalLayout.addWidget(self.tw_converters)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.pb_add_converter = QPushButton(converters)
        self.pb_add_converter.setObjectName(u"pb_add_converter")

        self.horizontalLayout.addWidget(self.pb_add_converter)

        self.pb_modify_converter = QPushButton(converters)
        self.pb_modify_converter.setObjectName(u"pb_modify_converter")

        self.horizontalLayout.addWidget(self.pb_modify_converter)

        self.pb_delete_converter = QPushButton(converters)
        self.pb_delete_converter.setObjectName(u"pb_delete_converter")

        self.horizontalLayout.addWidget(self.pb_delete_converter)

        self.pb_load_from_file = QPushButton(converters)
        self.pb_load_from_file.setObjectName(u"pb_load_from_file")

        self.horizontalLayout.addWidget(self.pb_load_from_file)

        self.pb_load_from_repo = QPushButton(converters)
        self.pb_load_from_repo.setObjectName(u"pb_load_from_repo")

        self.horizontalLayout.addWidget(self.pb_load_from_repo)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_2 = QLabel(converters)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_3.addWidget(self.label_2)

        self.le_converter_name = QLineEdit(converters)
        self.le_converter_name.setObjectName(u"le_converter_name")

        self.horizontalLayout_3.addWidget(self.le_converter_name)

        self.horizontalSpacer_3 = QSpacerItem(10, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_3)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label_3 = QLabel(converters)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_5.addWidget(self.label_3)

        self.le_converter_description = QLineEdit(converters)
        self.le_converter_description.setObjectName(u"le_converter_description")

        self.horizontalLayout_5.addWidget(self.le_converter_description)

        self.horizontalSpacer_4 = QSpacerItem(10, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_4)


        self.verticalLayout.addLayout(self.horizontalLayout_5)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.label = QLabel(converters)
        self.label.setObjectName(u"label")

        self.verticalLayout_3.addWidget(self.label)

        self.pb_code_help = QPushButton(converters)
        self.pb_code_help.setObjectName(u"pb_code_help")

        self.verticalLayout_3.addWidget(self.pb_code_help)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_3.addItem(self.verticalSpacer_2)


        self.horizontalLayout_2.addLayout(self.verticalLayout_3)

        self.pteCode = QPlainTextEdit(converters)
        self.pteCode.setObjectName(u"pteCode")
        font = QFont()
        font.setFamilies([u"Monospace"])
        self.pteCode.setFont(font)

        self.horizontalLayout_2.addWidget(self.pteCode)

        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.pb_save_converter = QPushButton(converters)
        self.pb_save_converter.setObjectName(u"pb_save_converter")

        self.verticalLayout_2.addWidget(self.pb_save_converter)

        self.pb_cancel_converter = QPushButton(converters)
        self.pb_cancel_converter.setObjectName(u"pb_cancel_converter")

        self.verticalLayout_2.addWidget(self.pb_cancel_converter)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)


        self.horizontalLayout_2.addLayout(self.verticalLayout_2)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_2)

        self.pb_cancel_widget = QPushButton(converters)
        self.pb_cancel_widget.setObjectName(u"pb_cancel_widget")

        self.horizontalLayout_4.addWidget(self.pb_cancel_widget)

        self.pbOK = QPushButton(converters)
        self.pbOK.setObjectName(u"pbOK")

        self.horizontalLayout_4.addWidget(self.pbOK)


        self.verticalLayout.addLayout(self.horizontalLayout_4)


        self.retranslateUi(converters)

        QMetaObject.connectSlotsByName(converters)
    # setupUi

    def retranslateUi(self, converters):
        converters.setWindowTitle(QCoreApplication.translate("converters", u"Time converters", None))
        self.label_4.setText(QCoreApplication.translate("converters", u"Converters", None))
        ___qtablewidgetitem = self.tw_converters.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("converters", u"Name", None));
        ___qtablewidgetitem1 = self.tw_converters.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("converters", u"Description", None));
        ___qtablewidgetitem2 = self.tw_converters.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("converters", u"Code", None));
        self.pb_add_converter.setText(QCoreApplication.translate("converters", u"Add new converter", None))
        self.pb_modify_converter.setText(QCoreApplication.translate("converters", u"Modify converter", None))
        self.pb_delete_converter.setText(QCoreApplication.translate("converters", u"Delete converter", None))
        self.pb_load_from_file.setText(QCoreApplication.translate("converters", u"Load converters from file", None))
        self.pb_load_from_repo.setText(QCoreApplication.translate("converters", u"Load converters from BORIS repository", None))
        self.label_2.setText(QCoreApplication.translate("converters", u"Name", None))
        self.label_3.setText(QCoreApplication.translate("converters", u"Description", None))
        self.label.setText(QCoreApplication.translate("converters", u"Python code", None))
        self.pb_code_help.setText(QCoreApplication.translate("converters", u"Help", None))
        self.pb_save_converter.setText(QCoreApplication.translate("converters", u"Save converter", None))
        self.pb_cancel_converter.setText(QCoreApplication.translate("converters", u"Cancel", None))
        self.pb_cancel_widget.setText(QCoreApplication.translate("converters", u"Cancel", None))
        self.pbOK.setText(QCoreApplication.translate("converters", u"OK", None))
    # retranslateUi




"""
BORIS plugin

number of occurences of behaviors
"""

import pandas as pd

__version__ = "0.0.1"
__version_date__ = "2025-04-10"
__plugin_name__ = "Behavior latency"
__author__ = "Olivier Friard - University of Torino - Italy"


import itertools


def run(df: pd.DataFrame):
    """
    Latency of a behavior after another.
    """

    df["start_time"] = pd.to_datetime(df["Start (s)"])
    df["end_time"] = pd.to_datetime(df["Stop (s)"])

    latency_by_subject: dict = {}

    for subject, group in df.groupby("subject"):
        behaviors = group["behavior"].tolist()
        # combinations = []
        # Utiliser itertools pour créer des combinaisons 2 à 2 des comportements
        for comb in itertools.combinations(behaviors, 2):
            # combinations.append(comb)

            last_A_end_time = None

            # Liste pour stocker les latences de chaque sujet
            subject_latency = []

            for index, row in group.iterrows():
                if row["behavior"] == comb[0]:
                    # Si on rencontre un comportement A, on réinitialise le temps de fin du comportement A
                    last_A_end_time = row["end_time"]
                    subject_latency.append(None)  # Pas de latence pour A
                elif row["behavior"] == comb[1] and last_A_end_time is not None:
                    # Si on rencontre un comportement B et qu'on a déjà vu un A avant
                    latency_time = row["start_time"] - last_A_end_time
                    subject_latency.append(latency_time)
                else:
                    # Si on rencontre un B mais sans A avant
                    subject_latency.append(None)

            # Ajout des latences calculées au DataFrame
            df.loc[group.index, f"latency {comb[1]} after {comb[0]}"] = subject_latency

            # Calcul de la latence totale ou moyenne par sujet
            latency_by_subject[(subject, comb)] = df.groupby("subject")["latency"].agg(["sum", "mean"])

    return str(latency_by_subject)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

from decimal import Decimal as dec
import logging

from PySide6.QtWidgets import (
    QDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
)

from . import config as cfg
from . import dialog
from .edit_event_ui import Ui_Form


class DlgEditEvent(QDialog, Ui_Form):
    def __init__(
        self,
        observation_type: str,
        time_value: dec = dec("NaN"),
        image_idx=None,
        current_time=0,
        time_format: str = cfg.S,
        show_set_current_time: bool = False,
        exif_date_time: dec | None = None,
        parent=None,
    ):
        super().__init__(parent)
        self.setupUi(self)
        self.time_value = time_value
        self.image_idx = image_idx
        self.observation_type = observation_type

        self.pb_set_to_current_time.setVisible(show_set_current_time)
        self.current_time = current_time
        self.exif_date_time = exif_date_time

        # hide frame index for all observations
        # frame index is determined in base of time
        for w in (
            self.lb_frame_idx,
            self.sb_frame_idx,
            self.cb_set_frame_idx_na,
        ):
            w.setVisible(False)

        # hide image index
        if observation_type in (cfg.LIVE, cfg.MEDIA):
            # hide image index
            for w in (
                self.cb_set_time_na,
                self.gb_image_index,
            ):
                w.setVisible(False)

        # widget for time
        self.time_widget = dialog.get_time_widget(self.time_value)

        if time_format == cfg.S:
            self.time_widget.rb_seconds.setChecked(True)
        if time_format == cfg.HHMMSS:
            self.time_widget.rb_time.setChecked(True)
        if not self.time_value.is_nan() and int(self.time_value) > cfg.DATE_CUTOFF:
            self.time_widget.rb_datetime.setChecked(True)

        self.horizontalLayout_3.insertWidget(0, self.time_widget)

        if observation_type == cfg.IMAGES:
            # hide frame index widgets
            self.pb_set_to_current_time.setVisible(self.exif_date_time is not None)
            self.sb_image_idx.setValue(self.image_idx)

        self.pb_set_to_current_time.clicked.connect(self.set_to_current_time)
        self.pb_set_to_current_image_index.clicked.connect(self.set_to_current_image_index)

        self.cb_set_time_na.stateChanged.connect(self.time_na)

        # self.cb_set_frame_idx_na.stateChanged.connect(self.frame_idx_na)
        self.pbOK.clicked.connect(self.close_widget)
        self.pbCancel.clicked.connect(self.reject)

    def close_widget(self):
        """
        close the widget
        """
        if self.observation_type in (cfg.IMAGES):
            if self.sb_image_idx.value() == 0:
                QMessageBox.warning(self, cfg.programName, "The image index cannot be null")
                return
        self.accept()

    def set_to_current_image_index(self):
        """
        set image index to current image index
        """
        if self.observation_type in (cfg.IMAGES):
            self.sb_image_idx.setValue(int(self.current_time))

    def set_to_current_time(self):
        """
        set time to current media time
        """

        print(f"{self.current_time=}")

        if self.observation_type in (cfg.LIVE, cfg.MEDIA):
            self.time_widget.set_time(dec(float(self.current_time)))

        if self.observation_type == cfg.IMAGES:
            if self.exif_date_time is not None:
                self.time_widget.set_time(dec(self.exif_date_time))

    # def frame_idx_na(self):
    #     """
    #     set/unset frame index NA
    #     """
    #     self.lb_frame_idx.setEnabled(not self.cb_set_frame_idx_na.isChecked())
    #     self.sb_frame_idx.setEnabled(not self.cb_set_frame_idx_na.isChecked())

    def time_na(self):
        """
        set/unset time to NA
        """

        logging.debug("time_na function")

        self.time_widget.setVisible(not self.cb_set_time_na.isChecked())
        self.time_widget.setEnabled(not self.cb_set_time_na.isChecked())

        self.pb_set_to_current_time.setVisible(not self.cb_set_time_na.isChecked() and self.exif_date_time is not None)
        self.pb_set_to_current_time.setEnabled(not self.cb_set_time_na.isChecked())


class EditSelectedEvents(QDialog):
    """
    "edit selected events" dialog box
    """

    def __init__(self):
        super(EditSelectedEvents, self).__init__()

        self.setWindowTitle("Edit selected events")

        hbox = QVBoxLayout(self)

        self.rbSubject = QRadioButton("Subject")
        self.rbSubject.setChecked(False)
        self.rbSubject.toggled.connect(self.rb_changed)
        hbox.addWidget(self.rbSubject)

        self.rbBehavior = QRadioButton("Behavior")
        self.rbBehavior.setChecked(False)
        self.rbBehavior.toggled.connect(self.rb_changed)
        hbox.addWidget(self.rbBehavior)

        self.lb = QLabel("New value")
        hbox.addWidget(self.lb)
        self.newText = QListWidget(self)
        hbox.addWidget(self.newText)

        self.rbComment = QRadioButton("Comment")
        self.rbComment.setChecked(False)
        self.rbComment.toggled.connect(self.rb_changed)
        hbox.addWidget(self.rbComment)

        self.lbComment = QLabel("New comment")
        hbox.addWidget(self.lbComment)

        self.commentText = QLineEdit()
        hbox.addWidget(self.commentText)

        hbox2 = QHBoxLayout(self)
        self.pbOK = QPushButton("OK")
        self.pbOK.clicked.connect(self.pbOK_clicked)
        self.pbCancel = QPushButton("Cancel")
        self.pbCancel.clicked.connect(self.pbCancel_clicked)
        hbox2.addWidget(self.pbCancel)
        hbox2.addWidget(self.pbOK)
        hbox.addLayout(hbox2)

        self.setLayout(hbox)

    def rb_changed(self):
        self.newText.setEnabled(not self.rbComment.isChecked())
        self.commentText.setEnabled(self.rbComment.isChecked())

        if self.rbSubject.isChecked():
            self.newText.clear()
            self.newText.addItems(self.all_subjects)

        if self.rbBehavior.isChecked():
            self.newText.clear()
            self.newText.addItems(self.all_behaviors)

        if self.rbComment.isChecked():
            self.newText.clear()

    def pbOK_clicked(self):
        """
        if not self.rbSubject.isChecked() and not self.rbBehavior.isChecked() and not self.rbComment.isChecked():
            QMessageBox.warning(
                None,
                cfg.programName,
                "You must select a field to be edited",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return
        """

        if (self.rbSubject.isChecked() or self.rbBehavior.isChecked()) and self.newText.selectedItems() == []:
            QMessageBox.warning(
                None,
                cfg.programName,
                "You must select a new value from the list",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        self.accept()

    def pbCancel_clicked(self):
        """
        Cancel editing
        """
        self.reject()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import importlib
import logging
import numpy as np
import pandas as pd
from pathlib import Path

from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMessageBox

from . import config as cfg
from . import project_functions
from . import dialog
from . import view_df


def add_plugins_to_menu(self):
    """
    add plugins to the plugins menu
    """
    for plugin_name in self.config_param.get(cfg.ANALYSIS_PLUGINS, {}):
        logging.debug(f"adding plugin '{plugin_name}' to menu")
        # Create an action for each submenu option
        action = QAction(self, triggered=lambda checked=False, name=plugin_name: run_plugin(self, name))
        action.setText(plugin_name)

        self.menu_plugins.addAction(action)


def get_plugin_name(plugin_path: str):
    """
    get name of plugin
    """
    # search plugin name
    plugin_name = None
    with open(plugin_path, "r") as f_in:
        for line in f_in:
            if line.startswith("__plugin_name__"):
                plugin_name = line.split("=")[1].strip().replace('"', "")
                break
    return plugin_name


def load_plugins(self):
    """
    load selected plugins in analysis menu
    """

    def msg():
        QMessageBox.warning(
            self,
            cfg.programName,
            f"A plugin with the same name is already loaded ({self.config_param[cfg.ANALYSIS_PLUGINS][plugin_name]}).\n\nThe plugin from {file_} is not loaded.",
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )

    self.menu_plugins.clear()
    self.config_param[cfg.ANALYSIS_PLUGINS] = {}

    # load BORIS plugins
    for file_ in sorted((Path(__file__).parent / "analysis_plugins").glob("*.py")):
        if file_.name == "__init__.py":
            continue
        if file_.name.startswith("_"):
            continue
        plugin_name = get_plugin_name(file_)
        if plugin_name is not None and plugin_name not in self.config_param.get(cfg.EXCLUDED_PLUGINS, set()):
            # check if plugin with same name already loaded
            if plugin_name in self.config_param[cfg.ANALYSIS_PLUGINS]:
                msg()
                continue

            self.config_param[cfg.ANALYSIS_PLUGINS][plugin_name] = str(file_)

    # load personal plugins
    if self.config_param.get(cfg.PERSONAL_PLUGINS_DIR, ""):
        for file_ in sorted(Path(self.config_param.get(cfg.PERSONAL_PLUGINS_DIR, "")).glob("*.py")):
            if file_.name == "__init__.py":
                continue
            if file_.name.startswith("_"):
                continue
            plugin_name = get_plugin_name(file_)
            if plugin_name is not None and plugin_name not in self.config_param.get(cfg.EXCLUDED_PLUGINS, set()):
                # check if plugin with same name already loaded
                if plugin_name in self.config_param[cfg.ANALYSIS_PLUGINS]:
                    msg()
                    continue

                self.config_param[cfg.ANALYSIS_PLUGINS][plugin_name] = str(file_)

    logging.debug(f"{self.config_param.get(cfg.ANALYSIS_PLUGINS, {})=}")


def plugin_df_filter(df: pd.DataFrame, observations_list: list = [], parameters: dict = {}) -> pd.DataFrame:
    """
    filter the dataframe following parameters

    filter by selected observations.
    filter by selected subjects.
    filter by selected behaviors.
    filter by time interval.
    """

    # filter selected observations
    df = df[df["Observation id"].isin(observations_list)]

    if parameters:
        # filter selected subjects
        df = df[df["Subject"].isin(parameters["selected subjects"])]

        # filter selected behaviors
        df = df[df["Behavior"].isin(parameters["selected behaviors"])]

        if parameters["time"] == cfg.TIME_OBS_INTERVAL:
            # filter each observation with observation interval start/stop

            # keep events between observation interval start time and observation interval stop/end
            df_interval = df[
                (
                    ((df["Start (s)"] >= df["Observation interval start"]) & (df["Start (s)"] <= df["Observation interval stop"]))
                    | ((df["Stop (s)"] >= df["Observation interval start"]) & (df["Stop (s)"] <= df["Observation interval stop"]))
                )
                | ((df["Start (s)"] < df["Observation interval start"]) & (df["Stop (s)"] > df["Observation interval stop"]))
            ]

            df_interval.loc[df["Start (s)"] < df["Observation interval start"], "Start (s)"] = df["Observation interval start"]
            df_interval.loc[df["Stop (s)"] > df["Observation interval stop"], "Stop (s)"] = df["Observation interval stop"]

            df_interval.loc[:, "Duration (s)"] = (df_interval["Stop (s)"] - df_interval["Start (s)"]).replace(0, np.nan)

            df = df_interval

        else:
            # filter selected time interval
            if parameters["start time"] is not None and parameters["end time"] is not None:
                MIN_TIME = parameters["start time"]
                MAX_TIME = parameters["end time"]

                # keep events between start time and end_time
                df_interval = df[
                    (
                        ((df["Start (s)"] >= MIN_TIME) & (df["Start (s)"] <= MAX_TIME))
                        | ((df["Stop (s)"] >= MIN_TIME) & (df["Stop (s)"] <= MAX_TIME))
                    )
                    | ((df["Start (s)"] < MIN_TIME) & (df["Stop (s)"] > MAX_TIME))
                ]

                # cut state events to interval
                df_interval.loc[df["Start (s)"] < MIN_TIME, "Start (s)"] = MIN_TIME
                df_interval.loc[df["Stop (s)"] > MAX_TIME, "Stop (s)"] = MAX_TIME

                df_interval.loc[:, "Duration (s)"] = (df_interval["Stop (s)"] - df_interval["Start (s)"]).replace(0, np.nan)

                df = df_interval

    print("filtered")
    print("=" * 50)

    # print(f"{df=}")

    return df


def run_plugin(self, plugin_name):
    """
    run plugin
    """

    if not self.project:
        QMessageBox.warning(
            self,
            cfg.programName,
            "No observations found. Open a project first",
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    logging.debug(f"{self.config_param.get(cfg.ANALYSIS_PLUGINS, {})=}")

    if plugin_name not in self.config_param.get(cfg.ANALYSIS_PLUGINS, {}):
        QMessageBox.critical(self, cfg.programName, f"Plugin '{plugin_name}' not found")
        return

    plugin_path = self.config_param.get(cfg.ANALYSIS_PLUGINS, {})[plugin_name]

    logging.debug(f"{plugin_path=}")

    if not Path(plugin_path).is_file():
        QMessageBox.critical(self, cfg.programName, f"The plugin {plugin_path} was not found.")
        return

    logging.debug(f"run plugin from {plugin_path}")

    module_name = Path(plugin_path).stem

    spec = importlib.util.spec_from_file_location(module_name, plugin_path)
    plugin_module = importlib.util.module_from_spec(spec)

    logging.debug(f"{plugin_module=}")

    spec.loader.exec_module(plugin_module)

    logging.info(
        f"{plugin_module.__plugin_name__} loaded v.{getattr(plugin_module, '__version__')} v. {getattr(plugin_module, '__version_date__')}"
    )

    selected_observations, parameters = self.obs_param()
    if not selected_observations:
        return

    logging.info("preparing dtaaframe for plugin")

    df = project_functions.project2dataframe(self.pj, selected_observations)

    logging.info("done")

    """
    logging.debug("dataframe info")
    logging.debug(f"{df.info()}")
    logging.debug(f"{df.head()}")
    """

    # filter the dataframe with parameters
    logging.info("filtering dataframe for plugin")
    filtered_df = plugin_df_filter(df, observations_list=selected_observations, parameters=parameters)
    logging.info("done")

    plugin_results = plugin_module.run(filtered_df)
    # test if plugin_tests is a tuple: if not transform to tuple
    if not isinstance(plugin_results, tuple):
        plugin_results = tuple([plugin_results])

    self.plugin_visu: list = []
    for result in plugin_results:
        if isinstance(result, str):
            self.plugin_visu.append(dialog.Results_dialog())
            self.plugin_visu[-1].setWindowTitle(plugin_name)
            self.plugin_visu[-1].ptText.clear()
            self.plugin_visu[-1].ptText.appendPlainText(result)
            self.plugin_visu[-1].show()
        elif isinstance(result, pd.DataFrame):
            self.plugin_visu.append(view_df.View_df(plugin_name, f"{plugin_module.__version__} ({plugin_module.__version_date__})", result))
            self.plugin_visu[-1].show()
        else:
            # result is not str nor dataframe
            QMessageBox.critical(
                None,
                cfg.programName,
                (
                    f"Plugin returns an unknown object type: {type(result)}\n\n"
                    "Plugins must return str and/or Pandas Dataframes.\n"
                    "Check the plugin code."
                ),
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )



"""
BORIS
Behavioral Observation Research Interactive Software

Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

from .core import main

name = "BORIS"



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'video_equalizer.ui'
##
## Created by: Qt User Interface Compiler version 6.8.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QHBoxLayout, QLabel,
    QPushButton, QSizePolicy, QSlider, QSpacerItem,
    QVBoxLayout, QWidget)

class Ui_Equalizer(object):
    def setupUi(self, Equalizer):
        if not Equalizer.objectName():
            Equalizer.setObjectName(u"Equalizer")
        Equalizer.resize(388, 284)
        self.verticalLayout = QVBoxLayout(Equalizer)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label_6 = QLabel(Equalizer)
        self.label_6.setObjectName(u"label_6")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_6.sizePolicy().hasHeightForWidth())
        self.label_6.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        self.label_6.setFont(font)

        self.verticalLayout.addWidget(self.label_6)

        self.cb_player = QComboBox(Equalizer)
        self.cb_player.setObjectName(u"cb_player")

        self.verticalLayout.addWidget(self.cb_player)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(Equalizer)
        self.label.setObjectName(u"label")
        self.label.setMinimumSize(QSize(70, 0))
        self.label.setMaximumSize(QSize(70, 16777215))

        self.horizontalLayout.addWidget(self.label)

        self.hs_brightness = QSlider(Equalizer)
        self.hs_brightness.setObjectName(u"hs_brightness")
        self.hs_brightness.setMinimumSize(QSize(200, 0))
        self.hs_brightness.setMaximumSize(QSize(200, 16777215))
        self.hs_brightness.setMinimum(-100)
        self.hs_brightness.setValue(0)
        self.hs_brightness.setOrientation(Qt.Horizontal)

        self.horizontalLayout.addWidget(self.hs_brightness)

        self.lb_brightness = QLabel(Equalizer)
        self.lb_brightness.setObjectName(u"lb_brightness")
        self.lb_brightness.setMinimumSize(QSize(25, 0))
        self.lb_brightness.setMaximumSize(QSize(25, 16777215))

        self.horizontalLayout.addWidget(self.lb_brightness)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_2 = QLabel(Equalizer)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMinimumSize(QSize(70, 0))
        self.label_2.setMaximumSize(QSize(70, 16777215))

        self.horizontalLayout_2.addWidget(self.label_2)

        self.hs_contrast = QSlider(Equalizer)
        self.hs_contrast.setObjectName(u"hs_contrast")
        self.hs_contrast.setMinimumSize(QSize(200, 0))
        self.hs_contrast.setMaximumSize(QSize(200, 16777215))
        self.hs_contrast.setMinimum(-100)
        self.hs_contrast.setValue(0)
        self.hs_contrast.setOrientation(Qt.Horizontal)

        self.horizontalLayout_2.addWidget(self.hs_contrast)

        self.lb_contrast = QLabel(Equalizer)
        self.lb_contrast.setObjectName(u"lb_contrast")
        self.lb_contrast.setMinimumSize(QSize(25, 0))
        self.lb_contrast.setMaximumSize(QSize(25, 16777215))

        self.horizontalLayout_2.addWidget(self.lb_contrast)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_3 = QLabel(Equalizer)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setMinimumSize(QSize(70, 0))
        self.label_3.setMaximumSize(QSize(70, 16777215))

        self.horizontalLayout_3.addWidget(self.label_3)

        self.hs_saturation = QSlider(Equalizer)
        self.hs_saturation.setObjectName(u"hs_saturation")
        self.hs_saturation.setMinimumSize(QSize(200, 0))
        self.hs_saturation.setMaximumSize(QSize(200, 16777215))
        self.hs_saturation.setMinimum(-100)
        self.hs_saturation.setValue(0)
        self.hs_saturation.setOrientation(Qt.Horizontal)

        self.horizontalLayout_3.addWidget(self.hs_saturation)

        self.lb_saturation = QLabel(Equalizer)
        self.lb_saturation.setObjectName(u"lb_saturation")
        self.lb_saturation.setMinimumSize(QSize(25, 0))
        self.lb_saturation.setMaximumSize(QSize(25, 16777215))

        self.horizontalLayout_3.addWidget(self.lb_saturation)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_4 = QLabel(Equalizer)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMinimumSize(QSize(70, 0))
        self.label_4.setMaximumSize(QSize(70, 16777215))

        self.horizontalLayout_4.addWidget(self.label_4)

        self.hs_gamma = QSlider(Equalizer)
        self.hs_gamma.setObjectName(u"hs_gamma")
        self.hs_gamma.setMinimumSize(QSize(200, 0))
        self.hs_gamma.setMaximumSize(QSize(200, 16777215))
        self.hs_gamma.setMinimum(-100)
        self.hs_gamma.setValue(0)
        self.hs_gamma.setOrientation(Qt.Horizontal)

        self.horizontalLayout_4.addWidget(self.hs_gamma)

        self.lb_gamma = QLabel(Equalizer)
        self.lb_gamma.setObjectName(u"lb_gamma")
        self.lb_gamma.setMinimumSize(QSize(25, 0))
        self.lb_gamma.setMaximumSize(QSize(25, 16777215))

        self.horizontalLayout_4.addWidget(self.lb_gamma)


        self.verticalLayout.addLayout(self.horizontalLayout_4)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label_5 = QLabel(Equalizer)
        self.label_5.setObjectName(u"label_5")
        self.label_5.setMinimumSize(QSize(70, 0))
        self.label_5.setMaximumSize(QSize(70, 16777215))

        self.horizontalLayout_5.addWidget(self.label_5)

        self.hs_hue = QSlider(Equalizer)
        self.hs_hue.setObjectName(u"hs_hue")
        self.hs_hue.setMinimumSize(QSize(200, 0))
        self.hs_hue.setMaximumSize(QSize(200, 16777215))
        self.hs_hue.setMinimum(-100)
        self.hs_hue.setValue(0)
        self.hs_hue.setOrientation(Qt.Horizontal)

        self.horizontalLayout_5.addWidget(self.hs_hue)

        self.lb_hue = QLabel(Equalizer)
        self.lb_hue.setObjectName(u"lb_hue")
        self.lb_hue.setMinimumSize(QSize(25, 0))
        self.lb_hue.setMaximumSize(QSize(25, 16777215))

        self.horizontalLayout_5.addWidget(self.lb_hue)


        self.verticalLayout.addLayout(self.horizontalLayout_5)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.pb_reset_all = QPushButton(Equalizer)
        self.pb_reset_all.setObjectName(u"pb_reset_all")

        self.horizontalLayout_6.addWidget(self.pb_reset_all)

        self.pb_reset = QPushButton(Equalizer)
        self.pb_reset.setObjectName(u"pb_reset")

        self.horizontalLayout_6.addWidget(self.pb_reset)


        self.verticalLayout.addLayout(self.horizontalLayout_6)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer)

        self.pb_close = QPushButton(Equalizer)
        self.pb_close.setObjectName(u"pb_close")

        self.horizontalLayout_7.addWidget(self.pb_close)


        self.verticalLayout.addLayout(self.horizontalLayout_7)


        self.retranslateUi(Equalizer)

        QMetaObject.connectSlotsByName(Equalizer)
    # setupUi

    def retranslateUi(self, Equalizer):
        Equalizer.setWindowTitle(QCoreApplication.translate("Equalizer", u"Video equalizer", None))
        self.label_6.setText(QCoreApplication.translate("Equalizer", u"Video equalizer", None))
        self.label.setText(QCoreApplication.translate("Equalizer", u"Brightness", None))
        self.lb_brightness.setText(QCoreApplication.translate("Equalizer", u"0", None))
        self.label_2.setText(QCoreApplication.translate("Equalizer", u"Contrast", None))
        self.lb_contrast.setText(QCoreApplication.translate("Equalizer", u"0", None))
        self.label_3.setText(QCoreApplication.translate("Equalizer", u"Saturation", None))
        self.lb_saturation.setText(QCoreApplication.translate("Equalizer", u"0", None))
        self.label_4.setText(QCoreApplication.translate("Equalizer", u"Gamma", None))
        self.lb_gamma.setText(QCoreApplication.translate("Equalizer", u"0", None))
        self.label_5.setText(QCoreApplication.translate("Equalizer", u"Hue", None))
        self.lb_hue.setText(QCoreApplication.translate("Equalizer", u"0", None))
        self.pb_reset_all.setText(QCoreApplication.translate("Equalizer", u"Reset all players", None))
        self.pb_reset.setText(QCoreApplication.translate("Equalizer", u"Reset current player", None))
        self.pb_close.setText(QCoreApplication.translate("Equalizer", u"Close", None))
    # retranslateUi




"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import os
import pathlib as pl
from decimal import Decimal as dec
from io import StringIO
import pandas as pd
import time

try:
    import pyreadr

    flag_pyreadr_loaded = True
except ModuleNotFoundError:
    flag_pyreadr_loaded = False


import tablib
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QListWidget,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
    QApplication,
)

from . import config as cfg
from . import (
    db_functions,
    dialog,
    gui_utilities,
    observation_operations,
    project_functions,
    select_observations,
    select_subj_behav,
    time_budget_functions,
)
from . import utilities as util


class timeBudgetResults(QWidget):
    """
    class for displaying time budget results in new window
    a function for exporting data in TSV, CSV, XLS and ODS formats is implemented

    Args:
        pj (dict): BORIS project
    """

    def __init__(self, pj, config_param):
        super().__init__()

        self.pj = pj
        self.config_param = config_param

        hbox = QVBoxLayout(self)

        self.label = QLabel("")
        hbox.addWidget(self.label)

        self.lw = QListWidget()
        # self.lw.setEnabled(False)
        self.lw.setMaximumHeight(100)
        hbox.addWidget(self.lw)

        self.lbTotalObservedTime = QLabel("")
        hbox.addWidget(self.lbTotalObservedTime)

        # behaviors excluded from total time
        self.excluded_behaviors_list = QLabel("")
        hbox.addWidget(self.excluded_behaviors_list)

        self.twTB = QTableWidget()
        hbox.addWidget(self.twTB)

        hbox2 = QHBoxLayout()

        self.pbSave = QPushButton("Save results", clicked=self.pbSave_clicked)
        hbox2.addWidget(self.pbSave)

        spacerItem = QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        hbox2.addItem(spacerItem)

        self.pbClose = QPushButton(cfg.CLOSE, clicked=self.close_clicked)
        hbox2.addWidget(self.pbClose)

        hbox.addLayout(hbox2)

        self.setWindowTitle("Time budget")

    def close_clicked(self):
        """
        save geometry of widget and close it
        """
        gui_utilities.save_geometry(self, "time budget")
        self.close()

    def pbSave_clicked(self):
        """
        save time budget analysis results in TSV, CSV, ODS, XLS format
        """

        def complete(lst: list, max_: int) -> list:
            """
            complete list with empty string until len = max

            Args:
                lst (list): list to complete
                max_ (int): length of the returned list

            Returns:
                list: completed list
            """

            lst.extend([""] * (max_ - len(lst)))
            return lst

        logging.debug("save time budget results to file")

        file_formats = (cfg.TSV, cfg.CSV, cfg.ODS, cfg.XLSX, cfg.XLS, cfg.HTML, cfg.TEXT_FILE, cfg.PANDAS_DF, cfg.RDS)

        file_name, filter_ = QFileDialog().getSaveFileName(self, "Save Time budget analysis", "", ";;".join(file_formats))

        if not file_name:
            return

        # add correct file extension if not present
        if pl.Path(file_name).suffix != f".{cfg.FILE_NAME_SUFFIX[filter_]}":
            if cfg.FILE_NAME_SUFFIX[filter_] != "cli":
                file_name = str(pl.Path(file_name)) + "." + cfg.FILE_NAME_SUFFIX[filter_]
            else:
                file_name = str(pl.Path(file_name))
            # check if file with new extension already exists
            if pl.Path(file_name).is_file():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", [cfg.CANCEL, cfg.OVERWRITE])
                    == cfg.CANCEL
                ):
                    return

        rows: list = []

        header: list = ["Observation id", "Observation date", "Description"]
        # indep var labels
        header.extend([self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"] for idx in self.pj[cfg.INDEPENDENT_VARIABLES]])
        header.extend(["Time budget start", "Time budget stop", "Time budget duration"])

        for idx in range(self.twTB.columnCount()):
            header.append(self.twTB.horizontalHeaderItem(idx).text())
        rows.append(header)

        col1: list = []
        # add obs id
        if self.lw.count() == 1:
            col1.append(self.lw.item(0).text())
        else:
            col1.append("NA, observations grouped")

        # add obs date
        if self.lw.count() == 1:
            col1.append(self.pj[cfg.OBSERVATIONS][self.lw.item(0).text()].get("date", "").replace("T", " "))
        else:
            # TODO: check if date is the same for all selected obs
            col1.append("NA, observations grouped")

        # description
        if self.lw.count() == 1:
            col1.append(util.eol2space(self.pj[cfg.OBSERVATIONS][self.lw.item(0).text()].get(cfg.DESCRIPTION, "")))
        else:
            col1.append("NA, observations grouped")

        # indep var values
        for idx in self.pj.get(cfg.INDEPENDENT_VARIABLES, []):
            if self.lw.count() == 1:
                # var has value in obs?
                if self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"] in self.pj[cfg.OBSERVATIONS][self.lw.item(0).text()].get(
                    cfg.INDEPENDENT_VARIABLES, []
                ):
                    col1.append(
                        self.pj[cfg.OBSERVATIONS][self.lw.item(0).text()][cfg.INDEPENDENT_VARIABLES][
                            self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"]
                        ]
                    )
                else:
                    col1.append("")
            else:
                # TODO: check if var value is the same for all selected obs
                col1.append("NA, observations grouped")

        if self.time_interval == cfg.TIME_ARBITRARY_INTERVAL:
            col1.extend([f"{self.min_time:0.3f}", f"{self.max_time:0.3f}", f"{self.max_time - self.min_time:0.3f}"])

        if self.time_interval == cfg.TIME_FULL_OBS:
            col1.extend(["Full observation", "Full observation", "Full observation"])

        if self.time_interval == cfg.TIME_EVENTS:
            col1.extend(["Limited to coded events", "Limited to coded events", "Limited to coded events"])

        for row_idx in range(self.twTB.rowCount()):
            values = []
            for col_idx in range(self.twTB.columnCount()):
                values.append(util.intfloatstr(self.twTB.item(row_idx, col_idx).text()))
            rows.append(col1 + values)

        """
        else:
            # observations list
            obs_header = ["Observation id", "Observation date", "Description"]

            # indep var
            obs_header.extend(
                [self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"] for idx in self.pj[cfg.INDEPENDENT_VARIABLES]]
            )

            obs_header.extend(["Time budget start", "Time budget stop", "Time budget duration"])

            obs_rows = []
            obs_rows.append(obs_header)
            for idx in range(self.lw.count()):
                row = []
                # obs id
                row.append(self.lw.item(idx).text())
                row.append(self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()].get("date", ""))
                row.append(util.eol2space(self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()].get(cfg.DESCRIPTION, "")))

                for idx2 in self.pj.get(cfg.INDEPENDENT_VARIABLES, []):
                    # var has value in obs?
                    if self.pj[cfg.INDEPENDENT_VARIABLES][idx2]["label"] in self.pj[cfg.OBSERVATIONS][
                        self.lw.item(idx).text()
                    ].get(cfg.INDEPENDENT_VARIABLES, []):
                        row.append(
                            self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()][cfg.INDEPENDENT_VARIABLES][
                                self.pj[cfg.INDEPENDENT_VARIABLES][idx2]["label"]
                            ]
                        )
                    else:
                        row.append("")
                # start stop duration
                row.extend([f"{self.min_time:0.3f}", f"{self.max_time:0.3f}", f"{self.max_time - self.min_time:0.3f}"])

                obs_rows.append(row)

            # write file with observations information
            data = tablib.Dataset()
            data.title = "Time budget - Observations information"

            for row in obs_rows:
                data.append(complete(row, max([len(r) for r in obs_rows])))

            with open(pl.Path(file_name).with_suffix(f".observations_info.{cfg.FILE_NAME_SUFFIX[filter_]}"), "wb") as f:
                if filter_ in [cfg.TSV, cfg.CSV, cfg.HTML]:
                    f.write(str.encode(data.export(cfg.FILE_NAME_SUFFIX[filter_])))
                if filter_ in [cfg.ODS, cfg.XLSX, cfg.XLS]:
                    f.write(data.export(cfg.FILE_NAME_SUFFIX[filter_]))
        """

        """
            rows.append(["Observations:"])
            for idx in range(self.lw.count()):
                rows.append([""])
                rows.append(["Observation id", self.lw.item(idx).text()])
                rows.append(["Observation date", self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()].get("date", "")])
                rows.append(
                    [
                        "Description",
                        util.eol2space(self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()].get(cfg.DESCRIPTION, "")),
                    ]
                )

                if cfg.INDEPENDENT_VARIABLES in self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()]:
                    rows.append(["Independent variables:"])
                    for var in self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()][cfg.INDEPENDENT_VARIABLES]:
                        rows.append(
                            [var, self.pj[cfg.OBSERVATIONS][self.lw.item(idx).text()][cfg.INDEPENDENT_VARIABLES][var]]
                        )

            if self.excluded_behaviors_list.text():
                s1, s2 = self.excluded_behaviors_list.text().split(": ")
                rows.extend([[""], [s1] + s2.split(", ")])

            rows.extend([[""], [""], ["Time budget:"]])


            # write header

            header = [self.twTB.horizontalHeaderItem(col_idx).text() for col_idx in range(self.twTB.columnCount())]
            rows.append(header)

            for row in range(self.twTB.rowCount()):
                values = []
                for col_idx in range(self.twTB.columnCount()):
                    values.append(util.intfloatstr(self.twTB.item(row, col_idx).text()))
                rows.append(values)
        """

        data = tablib.Dataset()
        data.title = "Time budget"

        for row in rows:
            data.append(complete(row, max([len(r) for r in rows])))

        if filter_ in (cfg.PANDAS_DF, cfg.RDS):
            pass

            # build pandas dataframe from the tsv export of tablib dataset
            dtype = {
                "Observation id": str,
                "Observation date": str,
                "Description": str,
                "Time budget start": str,
                "Time budget stop": str,
                "Time budget duration": str,
                "Subject": str,
                "Behavior": str,
                "Modifiers": str,
                "Total number of occurences": float,
                "Total duration (s)": float,
                "Duration mean (s)": float,
                "Duration std dev": float,
                "inter-event intervals mean (s)": float,
                "inter-event intervals std dev": float,
                "% of total length	": float,
            }

            # indep var values
            for idx in self.pj.get(cfg.INDEPENDENT_VARIABLES, []):
                if self.pj[cfg.INDEPENDENT_VARIABLES][idx]["type"] == "numeric":
                    dtype[self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"]] = float
                else:
                    dtype[self.pj[cfg.INDEPENDENT_VARIABLES][idx]["label"]] = str

            df = pd.read_csv(
                StringIO(data.export("tsv")),
                sep="\t",
                dtype=dtype,
                parse_dates=[1],
            )

            if filter_ == cfg.PANDAS_DF:
                df.to_pickle(file_name)

            if flag_pyreadr_loaded and filter_ == cfg.RDS:
                pyreadr.write_rds(file_name, df)

            return

        # write results
        with open(file_name, "wb") as f:
            if filter_ in (cfg.TSV, cfg.CSV, cfg.HTML, cfg.TEXT_FILE):
                f.write(str.encode(data.export(cfg.FILE_NAME_SUFFIX[filter_])))
            if filter_ in (cfg.ODS, cfg.XLSX, cfg.XLS):
                f.write(data.export(cfg.FILE_NAME_SUFFIX[filter_]))


def time_budget(self, mode: str, mode2: str = "list"):
    """
    time budget (by behavior or category)

    Args:
        mode (str): ["by_behavior", "by_category"]
        mode2 (str): must be in ["list", "current"]
                     "current" time budget of current observation
    """

    if mode2 == "current":
        if self.observationId:
            selected_observations = [self.observationId]
        else:
            return

    if mode2 == "list":
        _, selected_observations = select_observations.select_observations2(self, mode=cfg.MULTIPLE, windows_title="")

        if not selected_observations:
            return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    flagGroup: bool = False
    if len(selected_observations) > 1:
        flagGroup = (
            dialog.MessageDialog(cfg.programName, "Group the selected observations in a single time budget analysis?", [cfg.YES, cfg.NO])
            == cfg.YES
        )

    max_media_duration_all_obs, total_media_duration_all_obs = observation_operations.media_duration(
        self.pj[cfg.OBSERVATIONS], selected_observations
    )

    logging.debug(f"max_media_duration_all_obs: {max_media_duration_all_obs}, total_media_duration_all_obs={total_media_duration_all_obs}")

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)

    start_interval, end_interval = observation_operations.time_intervals_range(self.pj[cfg.OBSERVATIONS], selected_observations)

    parameters: dict = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=start_coding,
        end_coding=end_coding,
        # start_interval=start_interval,
        # end_interval=end_interval,
        start_interval=None,
        end_interval=None,
        maxTime=max_media_duration_all_obs,
        by_category=(mode == "by_category"),
        n_observations=len(selected_observations),
        show_exclude_non_coded_modifiers=True,
    )
    if parameters == {}:
        return

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    logging.debug(f"{parameters=}")

    # ask for excluding behaviors durations from total time
    if start_coding is not None and not start_coding.is_nan():
        cancel_pressed, parameters[cfg.EXCLUDED_BEHAVIORS] = self.filter_behaviors(
            title="Select behaviors to exclude from the total time",
            text=("The duration of the selected behaviors will be subtracted from the total time"),
            table="",
            behavior_type=cfg.STATE_EVENT_TYPES,
        )
        if cancel_pressed:
            return
    else:
        parameters[cfg.EXCLUDED_BEHAVIORS] = []

    self.statusbar.showMessage(f"Generating time budget for {len(selected_observations)} observation(s)")
    QApplication.processEvents()

    # check if time_budget window must be used
    if flagGroup or len(selected_observations) == 1:
        t0 = time.time()

        cursor = db_functions.load_events_in_db(
            self.pj,
            parameters[cfg.SELECTED_SUBJECTS],
            selected_observations,
            parameters[cfg.SELECTED_BEHAVIORS],
            time_interval=cfg.TIME_FULL_OBS,
        )

        total_observation_time = 0
        for obsId in selected_observations:
            obs_length = observation_operations.observation_total_length(self.pj[cfg.OBSERVATIONS][obsId])

            if obs_length == dec(-1):  # media length not available
                parameters[cfg.TIME_INTERVAL] = cfg.TIME_EVENTS

            if obs_length == dec(-2):  # images obs without time
                parameters[cfg.TIME_INTERVAL] = cfg.TIME_EVENTS

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_FULL_OBS:  # media file duration
                min_time = float(0)
                # check if the last event is recorded after media file length
                try:
                    if float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0]) > float(obs_length):
                        max_time = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0])
                    else:
                        max_time = float(obs_length)
                except Exception:
                    max_time = float(obs_length)

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_OBS_INTERVAL:
                obs_interval = self.pj[cfg.OBSERVATIONS][obsId].get(cfg.OBSERVATION_TIME_INTERVAL, [0, 0])
                offset = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.TIME_OFFSET])
                min_time = float(obs_interval[0]) + offset
                # Use max media duration for max time if no interval is defined (=0)
                max_time = float(obs_interval[1]) + offset if obs_interval[1] != 0 else float(obs_length)

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_EVENTS:  # events duration
                try:
                    min_time = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][0][0])  # first event
                except Exception:
                    min_time = float(0)
                try:
                    max_time = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0])  # last event
                except Exception:
                    # TODO: set to 0 if no events ?
                    max_time = float(obs_length)

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_ARBITRARY_INTERVAL:
                min_time = float(parameters[cfg.START_TIME])
                max_time = float(parameters[cfg.END_TIME])

                # check intervals
                for subj in parameters[cfg.SELECTED_SUBJECTS]:
                    for behav in parameters[cfg.SELECTED_BEHAVIORS]:
                        # if cfg.POINT in self.eventType(behav).upper():
                        if project_functions.event_type(behav, self.pj[cfg.ETHOGRAM]) in cfg.POINT_EVENT_TYPES:
                            continue

                        # extract modifiers

                        cursor.execute(
                            "SELECT distinct modifiers FROM events WHERE observation = ? AND subject = ? AND code = ?",
                            (obsId, subj, behav),
                        )
                        distinct_modifiers = list(cursor.fetchall())

                        # logging.debug("distinct_modifiers: {}".format(distinct_modifiers))

                        for modifier in distinct_modifiers:
                            # logging.debug("modifier #{}#".format(modifier[0]))

                            # insert events at boundaries of time interval
                            if (
                                len(
                                    cursor.execute(
                                        (
                                            "SELECT * FROM events "
                                            "WHERE observation = ? AND subject = ? AND code = ? AND modifiers = ? "
                                            "AND occurence < ?"
                                        ),
                                        (obsId, subj, behav, modifier[0], min_time),
                                    ).fetchall()
                                )
                                % 2
                            ):
                                cursor.execute(
                                    ("INSERT INTO events (observation, subject, code, type, modifiers, occurence) VALUES (?,?,?,?,?,?)"),
                                    (obsId, subj, behav, "STATE", modifier[0], min_time),
                                )

                            if (
                                len(
                                    cursor.execute(
                                        (
                                            "SELECT * FROM events WHERE observation = ? AND subject = ? AND code = ? "
                                            "AND modifiers = ? AND occurence > ?"
                                        ),
                                        (obsId, subj, behav, modifier[0], max_time),
                                    ).fetchall()
                                )
                                % 2
                            ):
                                cursor.execute(
                                    ("INSERT INTO events (observation, subject, code, type, modifiers, occurence) VALUES (?,?,?,?,?,?)"),
                                    (obsId, subj, behav, "STATE", modifier[0], max_time),
                                )
                        try:
                            cursor.execute("COMMIT")
                        except Exception:
                            pass

            total_observation_time += max_time - min_time

            # delete all events out of time interval from db
            cursor.execute(
                "DELETE FROM events WHERE observation = ? AND (occurence < ? OR occurence > ?)",
                (obsId, min_time, max_time),
            )
            try:
                cursor.execute("COMMIT")
            except Exception:
                pass

        out, categories = time_budget_functions.time_budget_analysis(
            self.pj[cfg.ETHOGRAM], cursor, selected_observations, parameters, by_category=(mode == "by_category")
        )

        # check excluded behaviors
        excl_behaviors_total_time = {}
        for element in out:
            if element["subject"] not in excl_behaviors_total_time:
                excl_behaviors_total_time[element["subject"]] = 0
            if element["behavior"] in parameters[cfg.EXCLUDED_BEHAVIORS]:
                excl_behaviors_total_time[element["subject"]] += element["duration"] if not isinstance(element["duration"], str) else 0

        # widget for results visualization
        self.tb = timeBudgetResults(self.pj, self.config_param)

        # add min and max time
        self.tb.time_interval = parameters[cfg.TIME_INTERVAL]
        self.tb.min_time = min_time
        self.tb.max_time = max_time

        # observations list
        self.tb.label.setText("Selected observations")
        for obs_id in selected_observations:
            # self.tb.lw.addItem(f"{obs_id}  {self.pj[OBSERVATIONS][obs_id]['date']}  {self.pj[OBSERVATIONS][obs_id]['description']}")
            self.tb.lw.addItem(obs_id)

        # media length
        if len(selected_observations) > 1:
            if total_observation_time:
                if self.timeFormat == cfg.HHMMSS:
                    self.tb.lbTotalObservedTime.setText(f"Total observation length: {util.seconds2time(total_observation_time)}")
                if self.timeFormat == cfg.S:
                    self.tb.lbTotalObservedTime.setText(f"Total observation length: {float(total_observation_time):0.3f}")
            else:
                self.tb.lbTotalObservedTime.setText("Total observation length: not available")
        else:
            if self.timeFormat == cfg.HHMMSS:
                self.tb.lbTotalObservedTime.setText(f"Analysis from {util.seconds2time(min_time)} to {util.seconds2time(max_time)}")
            if self.timeFormat == cfg.S:
                self.tb.lbTotalObservedTime.setText(f"Analysis from {float(min_time):0.3f} to {float(max_time):0.3f} s")

        # behaviors excluded from total time
        if parameters[cfg.EXCLUDED_BEHAVIORS]:
            self.tb.excluded_behaviors_list.setText(
                "Behaviors excluded from total time: " + (", ".join(parameters[cfg.EXCLUDED_BEHAVIORS]))
            )
        else:
            self.tb.excluded_behaviors_list.setVisible(False)

        self.statusbar.showMessage(f"Time budget generated in {round(time.time() - t0, 3)} s", 5000)
        logging.debug("Time budget generated")

        if mode == "by_behavior":
            tb_fields = [
                "Subject",
                "Behavior",
                "Modifiers",
                "Total number of occurences",
                "Total duration (s)",
                "Duration mean (s)",
                "Duration std dev",
                "inter-event intervals mean (s)",
                "inter-event intervals std dev",
                "% of total length",
            ]
            fields = [
                "subject",
                "behavior",
                "modifiers",
                "number",
                "duration",
                "duration_mean",
                "duration_stdev",
                "inter_duration_mean",
                "inter_duration_stdev",
            ]

            self.tb.twTB.setColumnCount(len(tb_fields))
            self.tb.twTB.setHorizontalHeaderLabels(tb_fields)

            for row in out:
                self.tb.twTB.setRowCount(self.tb.twTB.rowCount() + 1)
                column = 0
                for field in fields:
                    if isinstance(row[field], float):
                        item = QTableWidgetItem(f"{row[field]:.3f}")
                    else:
                        item = QTableWidgetItem(str(row[field]).replace(" ()", ""))
                    # no modif allowed
                    item.setFlags(Qt.ItemIsEnabled)
                    self.tb.twTB.setItem(self.tb.twTB.rowCount() - 1, column, item)
                    column += 1

                # % of total time
                if row["duration"] in (0, cfg.NA):
                    item = QTableWidgetItem(str(row["duration"]))
                elif row["duration"] not in ("-", cfg.UNPAIRED) and not start_coding.is_nan():
                    tot_time = float(total_observation_time)
                    # substract time of excluded behaviors from the total for the subject
                    if row["subject"] in excl_behaviors_total_time and row["behavior"] not in parameters[cfg.EXCLUDED_BEHAVIORS]:
                        tot_time -= excl_behaviors_total_time[row["subject"]]
                    item = QTableWidgetItem(f"{row['duration'] / tot_time * 100:.1f}" if tot_time > 0 else cfg.NA)

                else:
                    item = QTableWidgetItem("-")

                item.setFlags(Qt.ItemIsEnabled)
                self.tb.twTB.setItem(self.tb.twTB.rowCount() - 1, column, item)

        if mode == "by_category":
            tb_fields = ["Subject", "Category", "Total number", "Total duration (s)"]
            fields = ["number", "duration"]

            self.tb.twTB.setColumnCount(len(tb_fields))
            self.tb.twTB.setHorizontalHeaderLabels(tb_fields)

            for subject in categories:
                for category in categories[subject]:
                    self.tb.twTB.setRowCount(self.tb.twTB.rowCount() + 1)

                    column = 0
                    item = QTableWidgetItem(subject)
                    item.setFlags(Qt.ItemIsEnabled)
                    self.tb.twTB.setItem(self.tb.twTB.rowCount() - 1, column, item)

                    column = 1
                    if category == "":
                        item = QTableWidgetItem("No category")
                    else:
                        item = QTableWidgetItem(category)
                    item.setFlags(Qt.ItemIsEnabled)
                    self.tb.twTB.setItem(self.tb.twTB.rowCount() - 1, column, item)

                    for field in fields:
                        column += 1

                        if field == "duration":
                            try:
                                item = QTableWidgetItem(f"{categories[subject][category][field]:0.3f}")
                            except Exception:
                                item = QTableWidgetItem(categories[subject][category][field])
                        else:
                            item = QTableWidgetItem(str(categories[subject][category][field]))
                        item.setFlags(Qt.ItemIsEnabled)
                        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                        self.tb.twTB.setItem(self.tb.twTB.rowCount() - 1, column, item)

        self.tb.twTB.resizeColumnsToContents()

        gui_utilities.restore_geometry(self.tb, "time budget", (800, 600))

        self.tb.show()

    if not flagGroup and len(selected_observations) > 1:
        output_format, ok = QInputDialog.getItem(
            self,
            "Time budget analysis format",
            "Available formats",
            (
                cfg.TSV,
                cfg.CSV,
                cfg.ODS,
                cfg.ODS_WB,
                cfg.XLSX,
                cfg.XLSX_WB,
                cfg.HTML,
                cfg.XLS,
            ),
            0,
            False,
        )
        if not ok:
            return

        extension = cfg.FILE_NAME_SUFFIX[output_format]

        if output_format in (cfg.ODS_WB, cfg.XLSX_WB):
            workbook = tablib.Databook()

            wb_file_name, filter_ = QFileDialog(self).getSaveFileName(self, "Save Time budget analysis", "", output_format)
            if not wb_file_name:
                return

            if pl.Path(wb_file_name).suffix != f".{cfg.FILE_NAME_SUFFIX[filter_]}":
                wb_file_name = str(pl.Path(wb_file_name)) + "." + cfg.FILE_NAME_SUFFIX[filter_]
                # check if file with new extension already exists
                if pl.Path(wb_file_name).is_file():
                    if (
                        dialog.MessageDialog(cfg.programName, f"The file {wb_file_name} already exists.", [cfg.CANCEL, cfg.OVERWRITE])
                        == cfg.CANCEL
                    ):
                        return

        else:  # not workbook
            exportDir = QFileDialog(self).getExistingDirectory(
                self,
                "Choose a directory to save the time budget analysis",
                os.path.expanduser("~"),
                options=QFileDialog.ShowDirsOnly,
            )
            if not exportDir:
                return

        if mode == "by_behavior":
            tb_fields = [
                "Subject",
                "Behavior",
                "Modifiers",
                "Total number of occurences",
                "Total duration (s)",
                "Duration mean (s)",
                "Duration std dev",
                "inter-event intervals mean (s)",
                "inter-event intervals std dev",
                "% of total length",
            ]
            fields = [
                "subject",
                "behavior",
                "modifiers",
                "number",
                "duration",
                "duration_mean",
                "duration_stdev",
                "inter_duration_mean",
                "inter_duration_stdev",
            ]

        if mode == "by_category":
            tb_fields = ["Subject", "Category", "Total number of occurences", "Total duration (s)"]
            fields = ["subject", "category", "number", "duration"]

        mem_command = ""
        for obsId in selected_observations:
            cursor = db_functions.load_events_in_db(self.pj, parameters[cfg.SELECTED_SUBJECTS], [obsId], parameters[cfg.SELECTED_BEHAVIORS])

            obs_length = observation_operations.observation_total_length(self.pj[cfg.OBSERVATIONS][obsId])

            if obs_length == -1:
                obs_length = 0

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_FULL_OBS:
                min_time = float(0)
                # check if the last event is recorded after media file length
                try:
                    if float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0]) > float(obs_length):
                        max_time = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0])
                    else:
                        max_time = float(obs_length)
                except Exception:
                    max_time = float(obs_length)

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_EVENTS:
                try:
                    min_time = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][0][0])
                except Exception:
                    min_time = float(0)
                try:
                    max_time = float(self.pj[cfg.OBSERVATIONS][obsId][cfg.EVENTS][-1][0])
                except Exception:
                    max_time = float(obs_length)

            if parameters[cfg.TIME_INTERVAL] == cfg.TIME_ARBITRARY_INTERVAL:
                min_time = float(parameters[cfg.START_TIME])
                max_time = float(parameters[cfg.END_TIME])

                # check intervals
                for subj in parameters[cfg.SELECTED_SUBJECTS]:
                    for behav in parameters[cfg.SELECTED_BEHAVIORS]:
                        if project_functions.event_type(behav, self.pj[cfg.ETHOGRAM]) in cfg.POINT_EVENT_TYPES:
                            continue

                        cursor.execute(
                            "SELECT distinct modifiers FROM events WHERE observation = ? AND subject = ? AND code = ?",
                            (obsId, subj, behav),
                        )
                        distinct_modifiers = list(cursor.fetchall())

                        for modifier in distinct_modifiers:
                            if (
                                len(
                                    cursor.execute(
                                        (
                                            "SELECT * FROM events "
                                            "WHERE observation = ? AND subject = ? "
                                            "AND code = ? AND modifiers = ? AND occurence < ?"
                                        ),
                                        (obsId, subj, behav, modifier[0], min_time),
                                    ).fetchall()
                                )
                                % 2
                            ):
                                cursor.execute(
                                    ("INSERT INTO events (observation, subject, code, type, modifiers, occurence) VALUES (?,?,?,?,?,?)"),
                                    (obsId, subj, behav, "STATE", modifier[0], min_time),
                                )
                            if (
                                len(
                                    cursor.execute(
                                        (
                                            "SELECT * FROM events WHERE observation = ? AND subject = ? AND code = ?"
                                            " AND modifiers = ? AND occurence > ?"
                                        ),
                                        (obsId, subj, behav, modifier[0], max_time),
                                    ).fetchall()
                                )
                                % 2
                            ):
                                cursor.execute(
                                    ("INSERT INTO events (observation, subject, code, type, modifiers, occurence) VALUES (?,?,?,?,?,?)"),
                                    (obsId, subj, behav, cfg.STATE, modifier[0], max_time),
                                )
                        try:
                            cursor.execute("COMMIT")
                        except Exception:
                            pass

            cursor.execute(
                "DELETE FROM events WHERE observation = ? AND (occurence < ? OR occurence > ?)",
                (obsId, min_time, max_time),
            )

            out, categories = time_budget_functions.time_budget_analysis(
                self.pj[cfg.ETHOGRAM], cursor, [obsId], parameters, by_category=(mode == "by_category")
            )

            # check excluded behaviors
            excl_behaviors_total_time = {}
            for element in out:
                if element["subject"] not in excl_behaviors_total_time:
                    excl_behaviors_total_time[element["subject"]] = 0
                if element["behavior"] in parameters[cfg.EXCLUDED_BEHAVIORS]:
                    excl_behaviors_total_time[element["subject"]] += element["duration"] if element["duration"] != "NA" else 0

            rows: list = []
            col1: list = []
            # observation id
            col1.append(obsId)
            col1.append(self.pj[cfg.OBSERVATIONS][obsId].get("date", "").replace("T", ""))
            col1.append(util.eol2space(self.pj[cfg.OBSERVATIONS][obsId].get(cfg.DESCRIPTION, "")))
            header = ["Observation id", "Observation date", "Description"]

            indep_var_label: list = []
            indep_var_values: list = []
            for _, v in self.pj.get(cfg.INDEPENDENT_VARIABLES, {}).items():
                indep_var_label.append(v["label"])
                indep_var_values.append(self.pj[cfg.OBSERVATIONS][obsId].get(cfg.INDEPENDENT_VARIABLES, {}).get(v["label"], ""))

            header.extend(indep_var_label)
            col1.extend(indep_var_values)

            # interval analysis
            if dec(min_time).is_nan():  # check if observation has timestamp
                col1.extend([cfg.NA, cfg.NA, cfg.NA])
            else:
                col1.extend([f"{min_time:0.3f}", f"{max_time:0.3f}", f"{max_time - min_time:0.3f}"])
            header.extend(["Time budget start", "Time budget stop", "Time budget duration"])

            if mode == "by_behavior":
                # header
                rows.append(header + tb_fields)

                for row in out:
                    values = []
                    for field in fields:
                        values.append(str(row[field]).replace(" ()", ""))
                    # % of total time
                    if row["duration"] in (0, cfg.NA):
                        values.append(row["duration"])
                    elif row["duration"] not in ("-", cfg.UNPAIRED) and not start_coding.is_nan():
                        tot_time = float(max_time - min_time)
                        # substract duration of excluded behaviors from total time for each subject
                        if row["subject"] in excl_behaviors_total_time and row["behavior"] not in parameters[cfg.EXCLUDED_BEHAVIORS]:
                            tot_time -= excl_behaviors_total_time[row["subject"]]
                        # % of tot time
                        values.append(round(row["duration"] / tot_time * 100, 1) if tot_time > 0 else cfg.NA)
                    else:
                        values.append("-")

                    rows.append(col1 + values)

            if mode == "by_category":
                rows.append(header + tb_fields)

                for subject in categories:
                    for category in categories[subject]:
                        values = []
                        values.append(subject)
                        if category == "":
                            values.append("No category")
                        else:
                            values.append(category)

                        values.append(categories[subject][category]["number"])
                        try:
                            values.append(f"{categories[subject][category]['duration']:0.3f}")
                        except Exception:
                            values.append(categories[subject][category]["duration"])

                        rows.append(col1 + values)

            data = tablib.Dataset()
            data.title = obsId
            for row in rows:
                data.append(util.complete(row, max([len(r) for r in rows])))

            # check worksheet/workbook title for forbidden char (excel)
            data.title = util.safe_xl_worksheet_title(data.title, extension)

            if output_format in (cfg.ODS_WB, cfg.XLSX_WB):
                workbook.add_sheet(data)

            else:
                file_name = f"{pl.Path(exportDir) / pl.Path(util.safeFileName(obsId))}.{extension}"
                if mem_command != cfg.OVERWRITE_ALL and pl.Path(file_name).is_file():
                    if mem_command == "Skip all":
                        continue
                    mem_command = dialog.MessageDialog(
                        cfg.programName,
                        f"The file {file_name} already exists.",
                        [cfg.OVERWRITE, cfg.OVERWRITE_ALL, cfg.SKIP, cfg.SKIP_ALL, cfg.CANCEL],
                    )
                    if mem_command == cfg.CANCEL:
                        return
                    if mem_command in (cfg.SKIP, cfg.SKIP_ALL):
                        continue

                with open(file_name, "wb") as f:
                    if output_format in (cfg.TSV, cfg.CSV, cfg.HTML):
                        f.write(str.encode(data.export(cfg.FILE_NAME_SUFFIX[output_format])))

                    if output_format in (cfg.ODS, cfg.XLSX, cfg.XLS):
                        f.write(data.export(cfg.FILE_NAME_SUFFIX[output_format]))

        if output_format == cfg.XLSX_WB:
            with open(wb_file_name, "wb") as f:
                f.write(workbook.xlsx)
        if output_format == cfg.ODS_WB:
            with open(wb_file_name, "wb") as f:
                f.write(workbook.ods)



from .view_df_ui import Ui_Form
from PySide6.QtWidgets import QWidget, QFileDialog
from PySide6.QtCore import Qt, QAbstractTableModel

from . import config as cfg
from pathlib import Path
from . import dialog

try:
    import pyreadr

    flag_pyreadr_loaded = True
except ModuleNotFoundError:
    flag_pyreadr_loaded = False


class DataFrameModel(QAbstractTableModel):
    def __init__(self, dataframe):
        super().__init__()
        self._dataframe = dataframe

    def rowCount(self, parent=None):
        return self._dataframe.shape[0]

    def columnCount(self, parent=None):
        return self._dataframe.shape[1]

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            return str(self._dataframe.iat[index.row(), index.column()])
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self._dataframe.columns[section]
            elif orientation == Qt.Vertical:
                return self._dataframe.index[section]
        return None


class View_df(QWidget, Ui_Form):
    def __init__(self, plugin_name: str, plugin_version: str, df, parent=None):
        super().__init__()
        self.plugin_name = plugin_name
        self.df = df

        self.setupUi(self)
        self.lb_plugin_info.setText(f"{plugin_name} v. {plugin_version}")
        self.setWindowTitle(f"{plugin_name} v. {plugin_version}")

        # print(f"{self.df=}")

        self.pb_close.clicked.connect(self.close)
        self.pb_save.clicked.connect(self.save)

        model = DataFrameModel(self.df)
        self.tv_df.setModel(model)

    def save(self):
        file_formats = (
            cfg.TSV,
            cfg.CSV,
            cfg.ODS,
            cfg.XLSX,
            # cfg.XLS,
            cfg.HTML,
            # cfg.TBS,
            cfg.PANDAS_DF,
            cfg.RDS,
        )

        file_dialog_options = QFileDialog.Options()
        file_dialog_options |= QFileDialog.DontConfirmOverwrite

        file_name, filter_ = QFileDialog().getSaveFileName(
            None, f"Save {self.plugin_name}", "", ";;".join(file_formats), options=file_dialog_options
        )
        if not file_name:
            return

        outputFormat = cfg.FILE_NAME_SUFFIX[filter_]
        if Path(file_name).suffix != "." + outputFormat:
            file_name = f"{file_name}.{outputFormat}"
            if Path(file_name).exists():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", [cfg.CANCEL, cfg.OVERWRITE])
                    == cfg.CANCEL
                ):
                    return

        if filter_ == cfg.TSV:
            self.df.to_csv(file_name, sep="\t", index=False)
        if filter_ == cfg.CSV:
            self.df.to_csv(file_name, sep=";", index=False)
        if filter_ == cfg.XLSX:
            self.df.to_excel(file_name, index=False)
        if filter_ == cfg.ODS:
            self.df.to_excel(file_name, index=False, engine="odf")
        if filter_ == cfg.HTML:
            self.df.to_html(file_name, index=False)
        if filter_ == cfg.PANDAS_DF:
            self.df.to_pickle(file_name)

        if filter_ == cfg.RDS and flag_pyreadr_loaded:
            pyreadr.write_rds(file_name, self.df)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import os
import subprocess
import tempfile
from pathlib import Path

from PySide6.QtWidgets import QFileDialog, QMessageBox

from . import config as cfg
from . import dialog, export_observation, select_subj_behav
from . import select_observations


def behavioral_strings_analysis(strings, behav_seq_separator):
    """
    Analyze behavioral strings
    """

    rows = strings[:]
    sequences = []
    for row in rows:
        if behav_seq_separator:
            r = row.strip().split(behav_seq_separator)
        else:
            r = list(row.strip())

        sequences.append(r)

    # extract unique behaviors
    unique_behaviors = []
    for seq in sequences:
        for c in seq:
            if c not in unique_behaviors:
                unique_behaviors.append(c)

    unique_behaviors.sort()

    return sequences, unique_behaviors


def observed_transitions_matrix(sequences, behaviours, mode="frequency") -> str:
    """
    create the normalized matrix of observed transitions
    mode:
    * frequency:
    * number
    * frequencies_after_behaviors
    """

    logging.debug("function: observed_transitions_matrix")
    logging.debug(f"behaviours: {behaviours}")

    if "" in behaviours:
        behaviours.remove("")

    transitions = {}
    for behaviour in behaviours:
        if not behaviour:
            continue
        transitions[behaviour] = {}
        for behaviour2 in behaviours:
            transitions[behaviour][behaviour2] = 0

    for seq in sequences:
        for i in range(len(seq) - 1):
            if not seq[i]:
                continue
            if seq[i] in behaviours and seq[i + 1] in behaviours:
                transitions[seq[i]][seq[i + 1]] += 1

    transitions_total_number = sum([sum(transitions[x].values()) for x in transitions])

    if not transitions_total_number:
        return False

    out = "\t" + "\t".join(list(behaviours)) + "\n"
    for behaviour in behaviours:
        out += "{}\t".format(behaviour)
        for behaviour2 in behaviours:
            if mode == "frequency":
                out += "{}\t".format(transitions[behaviour][behaviour2] / transitions_total_number)
            elif mode == "number":
                out += "{}\t".format(transitions[behaviour][behaviour2])
            elif mode == "frequencies_after_behaviors":
                if sum(transitions[behaviour].values()):
                    out += "{}\t".format(transitions[behaviour][behaviour2] / sum(transitions[behaviour].values()))
                else:
                    out += "{}\t".format(transitions[behaviour][behaviour2])
        out = out[:-1] + "\n"

    return out


def create_transitions_gv_from_matrix(matrix, cutoff_all=0, cutoff_behavior=0, edge_label="percent_node") -> tuple:
    """
    create code for GraphViz
    matrix: matrix of frequency
    edge_label: (percent_node, fraction_node)
    return string containing graphviz code
    """

    behaviours = matrix.split("\n")[0].strip().split("\t")
    transitions = {}

    for row in matrix.split("\n")[1:]:
        if not row:
            continue

        transitions[row.split("\t")[0]] = {}
        for idx, r in enumerate(row.split("\t")[1:]):
            if "." in r:
                transitions[row.split("\t")[0]][behaviours[idx]] = float(r)
            else:
                transitions[row.split("\t")[0]][behaviours[idx]] = int(r)

    """transitions_total_number = sum([sum(transitions[x].values()) for x in transitions])"""

    out = "digraph G { \n"

    for behaviour1 in behaviours:
        for behaviour2 in behaviours:
            if behaviour1 not in transitions or behaviour2 not in transitions:
                return True, "Error: the file does not seem a transition matrix"
            if transitions[behaviour1][behaviour2]:
                if edge_label == "percent_node":
                    if transitions[behaviour1][behaviour2] > cutoff_all:
                        out += '"{behaviour1}" -> "{behaviour2}" [label="{label:0.3f}"];\n'.format(
                            behaviour1=behaviour1, behaviour2=behaviour2, label=transitions[behaviour1][behaviour2]
                        )

                if edge_label == "fraction_node":
                    transition_sum = sum(transitions[behaviour1].values())
                    if transitions[behaviour1][behaviour2] / transition_sum > cutoff_behavior:
                        out += """"{behaviour1}" -> "{behaviour2}" [label="{label}%"];\n""".format(
                            behaviour1=behaviour1,
                            behaviour2=behaviour2,
                            label=round(transitions[behaviour1][behaviour2] / transition_sum * 100, 1),
                        )

    out += "\n}"
    return False, out


def transitions_matrix(self, mode):
    """
    create transitions frequencies matrix with selected observations, subjects and behaviors
    mode:
    * frequency
    * number
    * frequencies_after_behaviors
    """
    logging.debug("flag transitions_matrix function")

    # ask user observations to analyze
    _, selected_observations = select_observations.select_observations2(
        self, cfg.MULTIPLE, windows_title="Select observations for transitions matrix"
    )

    if not selected_observations:
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        show_include_modifiers=True,
        show_exclude_non_coded_behaviors=False,
        n_observations=len(selected_observations),
    )

    if parameters == {}:
        return

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        return

    flagMulti = False
    if len(parameters[cfg.SELECTED_SUBJECTS]) == 1:
        file_name, _ = QFileDialog().getSaveFileName(
            None,
            "Create matrix of transitions " + mode,
            "",
            "Transitions matrix files (*.txt *.tsv);;All files (*)",
        )
        if not file_name:
            return
    else:
        exportDir = QFileDialog.getExistingDirectory(
            self,
            "Choose a directory to save the transitions matrices",
            str(Path.home()),
            options=QFileDialog.ShowDirsOnly,
        )
        if not exportDir:
            return
        flagMulti = True

    flag_overwrite_all = False
    for subject in parameters[cfg.SELECTED_SUBJECTS]:
        logging.debug(f"subjects: {subject}")

        strings_list = []
        for obs_id in selected_observations:
            strings_list.append(
                export_observation.events_to_behavioral_sequences(self.pj, obs_id, subject, parameters, self.behav_seq_separator)
            )

        sequences, observed_behaviors = behavioral_strings_analysis(strings_list, self.behav_seq_separator)

        observed_matrix = observed_transitions_matrix(
            sequences, sorted(list(set(observed_behaviors + parameters[cfg.SELECTED_BEHAVIORS]))), mode=mode
        )

        if not observed_matrix:
            QMessageBox.warning(self, cfg.programName, f"No transitions found for <b>{subject}</b>")
            continue

        logging.debug(f"observed_matrix {mode}:\n{observed_matrix}")

        if flagMulti:
            try:
                nf = f"{exportDir}{os.sep}{subject}_transitions_{mode}_matrix.tsv"

                if os.path.isfile(nf) and not flag_overwrite_all:
                    answer = dialog.MessageDialog(
                        cfg.programName,
                        f"A file with same name already exists.<br><b>{nf}</b>",
                        ["Overwrite", "Overwrite all", cfg.CANCEL],
                    )
                    if answer == cfg.CANCEL:
                        continue
                    if answer == "Overwrite all":
                        flag_overwrite_all = True

                with open(nf, "w") as outfile:
                    outfile.write(observed_matrix)
            except Exception:
                QMessageBox.critical(self, cfg.programName, f"The file {nf} can not be saved")
        else:
            try:
                with open(file_name, "w") as outfile:
                    outfile.write(observed_matrix)

            except Exception:
                QMessageBox.critical(self, cfg.programName, f"The file {file_name} can not be saved")


def transitions_dot_script():
    """
    create dot script (graphviz language) from transitions frequencies matrix
    """

    file_names, _ = QFileDialog().getOpenFileNames(
        None,
        "Select one or more transitions matrix files",
        "",
        "Transitions matrix files (*.txt *.tsv);;All files (*)",
    )

    out = ""

    for file_name in file_names:
        with open(file_name, "r") as infile:
            result, gv = create_transitions_gv_from_matrix(infile.read(), cutoff_all=0, cutoff_behavior=0, edge_label="percent_node")
            if result:
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    gv,
                )
                return

            try:
                with open(file_name + ".gv", "w") as file_out:
                    file_out.write(gv)
            except Exception:
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    ("Error saving the file"),
                )
                return

            out += f"<b>{file_name}.gv</b> created<br>"

    if out:
        QMessageBox.information(
            None,
            cfg.programName,
            (f"{out}<br><br>The DOT scripts can be used with the Graphviz package or WebGraphviz to generate diagram"),
        )


def transitions_flow_diagram():
    """
    create flow diagram with graphviz (if installed) from transitions matrix
    """

    # check if dot present in path
    result = subprocess.getoutput("dot -V")
    if "graphviz" not in result:
        QMessageBox.critical(
            None,
            cfg.programName,
            (
                "The GraphViz package is not installed.<br>"
                "The <b>dot</b> program was not found in the path.<br><br>"
                'Go to <a href="http://www.graphviz.org">'
                "http://www.graphviz.org</a> for information"
            ),
        )
        return

    file_names, _ = QFileDialog().getOpenFileNames(
        None,
        "Select one or more transitions matrix files",
        "",
        "Transitions matrix files (*.txt *.tsv);;All files (*)",
    )

    out = ""
    for file_name in file_names:
        with open(file_name, "r") as infile:
            result, gv = create_transitions_gv_from_matrix(infile.read(), cutoff_all=0, cutoff_behavior=0, edge_label="percent_node")
            if result:
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    gv,
                )
                return

            with open(tempfile.gettempdir() + os.sep + os.path.basename(file_name) + ".tmp.gv", "w") as f:
                f.write(gv)
            result = subprocess.getoutput(
                (f'dot -Tpng -o "{file_name}.png" ' f'"{tempfile.gettempdir() + os.sep + os.path.basename(file_name)}.tmp.gv"')
            )
            if not result:
                out += f"<b>{file_name}.png</b> created<br>"
            else:
                out += f"Problem with <b>{file_name}</b><br>"

    if out:
        QMessageBox.information(None, cfg.programName, out)



"""
BORIS plugin

number of occurences of behaviors by independent_variable
"""

import pandas as pd

__version__ = "0.3.0"
__version_date__ = "2025-03-17"
__plugin_name__ = "Number of occurences of behaviors by subject by independent_variable"
__author__ = "Olivier Friard - University of Torino - Italy"


def run(df: pd.DataFrame):
    """
    Calculate the number of occurrences of behaviors by subject and by independent_variable.

    This plugin returns a Pandas dataframe
    """

    df_results: df.DataFrame = (
        df.groupby(
            [
                "independent variable 'Weather'",
                "Subject",
                "Behavior",
            ]
        )["Behavior"]
        .count()
        .reset_index(name="number of occurences")
    )

    return df_results
